from . import models
from rest_framework import serializers
from django.db.models import Q
from engines.serializers import ClusterPKField
from data.models import DATASET_TYPE

class DataSourceSerializer(serializers.ModelSerializer):
    owner = serializers.ReadOnlyField(source='owner.username')
    class Meta:
        model = models.DataSource
        fields = '__all__'

class DataSourcePKField(serializers.PrimaryKeyRelatedField):
    def get_queryset(self):
        user = self.context['request'].user
        queryset = models.DataSource.objects.filter(Q(public=True) | Q(owner=user))
        return queryset

class DatasetSerializer(serializers.ModelSerializer):
    owner = serializers.ReadOnlyField(source='owner.username')
    class Meta:
        model = models.Dataset
        fields = '__all__'

class DatasetReadSerializer(serializers.ModelSerializer):
    owner = serializers.ReadOnlyField(source='owner.username')
    type_name = serializers.ReadOnlyField()
    def to_representation(self, obj):
        # get the original representation
        ret = super(DatasetReadSerializer, self).to_representation(obj)
        if obj.owner!=self.context['request'].user:
            ret.pop('uri')
            ret.pop('remedy_script')
        return ret
    class Meta:
        model = models.Dataset
        fields = '__all__'

class DatasetPKField(serializers.PrimaryKeyRelatedField):
    def get_queryset(self):
        user = self.context['request'].user
        queryset = models.Dataset.objects.filter(Q(public=True) | Q(owner=user))
        return queryset

class DataEngineSerializer(serializers.ModelSerializer):
    owner = serializers.ReadOnlyField(source='owner.username')
    class Meta:
        model = models.DataEngine
        fields = '__all__'
        
class DataEnginePKField(serializers.PrimaryKeyRelatedField):
    def get_queryset(self):
        user = self.context['request'].user
        queryset = models.DataEngine.objects.filter(Q(public=True) | Q(owner=user))
        return queryset

class DataEngineSpacePKField(serializers.PrimaryKeyRelatedField):
    def get_queryset(self):
        user = self.context['request'].user
        queryset = models.DataEngine.objects.filter(Q(public=True) | Q(owner=user))
        return queryset

class DataInstanceSerializer(serializers.ModelSerializer):
    owner = serializers.ReadOnlyField(source='owner.username')
    cluster = ClusterPKField()
    dataset=DatasetPKField()
    engine=DataEnginePKField()#filter out the space unselected engines
    uri_elected = serializers.ReadOnlyField()
    uri_alive = serializers.ReadOnlyField()
    uri_total = serializers.ReadOnlyField()
    status_name = serializers.ReadOnlyField()
    class Meta:
        model = models.DataInstance
        fields = '__all__'

class DataInstancePKField(serializers.PrimaryKeyRelatedField):
    def get_queryset(self):
        user = self.context['request'].user
        qs = models.DataInstance.objects.all()
        if not user.is_superuser:
            qs = qs.filter(owner=user)
        return qs

class DataInstanceOperationSerializer(serializers.ModelSerializer):
    target=DataInstancePKField()
    class Meta:
        model = models.DataInstanceOperation
        fields = '__all__'


from uuid import UUID
import time
from django.db import transaction
from django.db.models import Value as V
from django.db.models.functions import Concat
from django.db.models import Max
from django.utils.timezone import now
from django.dispatch import receiver
from django.db.models.signals import pre_save, post_save, pre_delete, post_delete
from clouds.signals import materialized, executed, monitored, destroyed, tidied, selected
from clouds.signals import tidy_operation, select_operation
from clouds import utils
from .import models
from clouds.models import InstanceBlueprint, Instance, INSTANCE_STATUS, InstanceOperation, INSTANCE_OPERATION, OPERATION_STATUS, Mount, Group, GroupOperation

from django.dispatch import Signal
scaled_out = Signal(providing_args=["instance","name"])

@receiver(scaled_out)
def log(sender,instance,name,**kwargs):
    print('SIGNAL INFO:', sender._meta.app_label, sender._meta.verbose_name, instance, name)

from clouds.models import Cloud, bootstraped

from .utils import remedy_scale_ambari_bootstrap
@receiver(bootstraped, sender=Cloud)
def bootstrap(sender,instance,**kwargs):
    from django.conf import settings
    blueprints=list(instance.instanceblueprint_set.filter(name__startswith='packone-bootstap-', public=False))
    for s in settings.PACKONE_STACK_VDF_URLS:
        stack, created=models.Stack.objects.get_or_create(
            name=s['name'],
            defaults={
                'meta': '{"vdf_url": "'+s['vdf']+'"}',
                'public': True,
                'owner': instance.owner
            }
        )
        scale, created=models.Scale.objects.get_or_create(
            name='packone.bootstrap.{}.{}'.format(stack.name,instance.name),
            defaults={
                'stack': stack,
                'owner': instance.owner, #TODO only cloud owner can bootstrap the cloud
                '_remedy_script': remedy_scale_ambari_bootstrap(stack.meta_data['vdf_url'])
            }
        )
        scale.init_blueprints.add(*blueprints)

@receiver(post_delete, sender=InstanceBlueprint)
def cleanup_scale(sender, instance, **kwargs):
    models.Scale.objects.filter(init_blueprints__isnull=True, step_blueprints__isnull=True).delete()

@receiver(executed, sender=InstanceOperation)
def import_stack_engines(sender,instance,**kwargs):
    if instance.status==OPERATION_STATUS.success.value and instance.target.image.name=='packone-bootstrap-master1' and 'ambari localhost:8080 cluster create' in instance.script:
        cluster=instance.target.group_set.first().cluster_set.first()
        cluster.import_engine()

@receiver(materialized, sender=Group)
@receiver(post_save, sender=models.Cluster)
def scale_out(sender,instance,**kwargs):
    if sender==models.Cluster:
        if not kwargs['created'] or instance.deleting: return
        instance.scale_one_step()
        return
    for cluster in instance.cluster_set.select_for_update():
        cluster.built_time=now()
        cluster.save()
        old_steps=cluster.get_ready_steps().exclude(pk=instance.pk).select_for_update()
        if old_steps.exists():
            old_hosts_script=utils.remedy_script_hosts_add('\n'.join([step.hosts for step in old_steps]))
            new_hosts_script=utils.remedy_script_hosts_add(instance.hosts)   
            #TODO only running remedyless cluster can be scaled-out
            for step in old_steps:
                step.remedy(new_hosts_script)
            instance.remedy(old_hosts_script,manual=False)
            if cluster.scale.remedy_script_scale_out:
                instance.remedy(cluster.scale.remedy_script_scale_out,manual=False)
        elif cluster.scale.remedy_script:
            instance.remedy(cluster.scale.remedy_script,manual=False)
        scaled_out.send(sender=models.Cluster, instance=cluster, name='scaled_out')

@receiver(destroyed, sender=Group)
@transaction.atomic
def scale_in_cluster(sender,instance,**kwargs):
    for cluster in instance.cluster_set.select_for_update():
        cluster.steps.remove(instance)
        if cluster.deleting:
            if not cluster.steps.all().exists():
                destroyed.send(sender=models.Cluster, instance=cluster, name='destroyed')
                cluster.delete()
        elif instance.hosts:
            cluster.update_remedy_script(
                utils.remedy_script_hosts_remove(instance.hosts),
                heading=True
            )
            scale_in_script=''
            for h in instance.hosts.split('\n'):
                if h.startswith('#'): continue
                scale_in_script+=cluster.scale.remedy_script_scale_in.format(hostname=h.split(' ')[-1])+'\n'
            cluster.update_remedy_script(
                scale_in_script,
                heading=True
            )
            cluster.remedy()

@receiver(monitored, sender=Group)
@receiver(tidied, sender=models.ClusterOperation)
@receiver(executed, sender=models.ClusterOperation)
def monitor_status(sender, instance, **kwargs):
    if sender==Group:
        if instance.deleting: return
        for cluster in instance.cluster_set.all():
            status=cluster.steps.all().aggregate(Max('status'))['status__max']
            models.Cluster.objects.filter(pk=cluster.pk).update(status=status)
            cluster.refresh_from_db()
            monitored.send(sender=models.Cluster, instance=cluster, name='monitored')
    else:
        if instance.status==OPERATION_STATUS.waiting.value: return
        else:
            instance.target.monitor()

post_save.connect(tidy_operation,sender=models.ClusterOperation)
monitored.connect(select_operation,sender=models.Cluster)
@receiver(selected, sender=models.ClusterOperation)
def execute_operation(sender,instance,**kwargs):
    instance.execute()

#TODO use threading join to reduce last singal check
@receiver(executed, sender=GroupOperation)
@transaction.atomic
def close_cluster_operation(sender, instance, **kwargs):
    for running_op in models.ClusterOperation.objects.select_for_update().filter(
        batch_uuid=instance.batch_uuid,
        started_time__isnull=False
    ):
        if not running_op.get_remain_oprations().exists():
            running_op.completed_time=now()
            running_op.status=running_op.get_status()
            running_op.save()
            executed.send(sender=models.ClusterOperation, instance=running_op, name='executed')

        # elif instance.operation==models.COMPONENT_OPERATION.stop.value:
        #     print(instance.target.stop())
        # elif instance.operation==models.COMPONENT_OPERATION.restart.value:
        #     raise Exception('not implemented yet')
        # else:
        #     raise Exception('illegal operation')
        # instance.completed_time=now()

@receiver(post_delete, sender=GroupOperation)
def purge_cluster_operation(sender, instance, **kwargs):
    c_op=models.ClusterOperation.objects.filter(batch_uuid=instance.batch_uuid).first()
    if c_op and not c_op.get_sub_operations().exists():
        c_op.delete()

# @receiver(post_save, sender=models.EngineOperation)
# def operate_engine(sender,instance,created,**kwargs):
#     if created:
#         if not instance.engine.enabled:
#             raise Exception('cannot operate disabled engine')
#         if instance.operation==models.COMPONENT_OPERATION.start.value:
#             instance.engine.start(instance.pilot)
#         elif instance.operation==models.COMPONENT_OPERATION.stop.value:
#             instance.engine.stop(instance.pilot)
#         instance.completed_time=now()
#         instance.save()


from django.urls import re_path, include
from rest_framework import routers
from . import views

ROUTER = routers.DefaultRouter()
ROUTER.register(r'clouds', views.CloudViewSet)
ROUTER.register(r'images', views.ImageViewSet, basename='image')
ROUTER.register(r'instancetemplates', views.InstanceTemplateViewSet)
ROUTER.register(r'instanceblueprints', views.InstanceBlueprintViewSet, basename='instanceblueprint')
ROUTER.register(r'instances', views.InstanceViewSet, basename='instance')
ROUTER.register(r'instanceoperations', views.InstanceOperationViewSet, basename='instanceoperation')
ROUTER.register(r'volumes', views.VolumeViewSet, basename='volume')
ROUTER.register(r'mounts', views.MountViewSet, basename='mount')

urlpatterns = [
    re_path(r'^', include(ROUTER.urls)),
    re_path(
        r'^instancetemplate-autocomplete/$',
        views.InstanceTemplateAutocompleteView.as_view(),
        name='instancetemplate-autocomplete',
    ),
    re_path(
        r'^image-autocomplete/$',
        views.ImageAutocompleteView.as_view(),
        name='image-autocomplete',
    ),
    re_path(
        r'^mountinstance-autocomplete/$',
        views.MountInstanceAutocompleteView.as_view(),
        name='mountinstance-autocomplete',
    ),
]



from dal import autocomplete
from rest_framework import viewsets
from rest_framework.permissions import IsAuthenticated, SAFE_METHODS
from django.db.models import Q
from . import models
from . import serializers
from . import permissions
from .utils import get_refer_GET_parameter

#TODO: dal can be walked around!!!
class InstanceTemplateAutocompleteView(autocomplete.Select2QuerySetView):
    def get_queryset(self):
        if not self.request.user.is_authenticated:
            return models.InstanceTemplate.objects.none()
        cloud_id = self.forwarded.get('cloud', None)
        if not cloud_id:
            return models.InstanceTemplate.objects.none()
        image_id = self.forwarded.get('image', None)
        image=models.Image.objects.get(pk=image_id)
        qs = models.InstanceTemplate.objects.filter(cloud=cloud_id,ram__gte=image.min_ram,disk__gte=image.min_disk,enabled=True).filter(
            Q(owner=self.request.user) | Q(public=True)
        ).order_by('id')
        if self.q:
            qs = qs.filter(name__istartswith=self.q)
        return qs

class ImageAutocompleteView(autocomplete.Select2QuerySetView):
    def get_queryset(self):
        if not self.request.user.is_authenticated:
            return models.Image.objects.none()
        cloud = self.forwarded.get('cloud', None)
        if not cloud:
            return models.Image.objects.none()
        qs = models.Image.objects.filter(cloud=cloud,enabled=True).filter(
            Q(owner=self.request.user) | Q(public=True)
        ).order_by('id')
        if self.q:
            qs = qs.filter(name__istartswith=self.q)
        return qs

class MountInstanceAutocompleteView(autocomplete.Select2QuerySetView):
    def get_queryset(self):
        if not self.request.user.is_authenticated:
            return models.Instance.objects.none()
        volume_id = self.forwarded.get('volume', None)
        if not volume_id:
            return models.Instance.objects.none()
        volume=models.Volume.objects.get(pk=volume_id)
        qs = models.Instance.objects.filter(
            cloud=volume.cloud,
            status__in=[
                models.INSTANCE_STATUS.poweroff.value,
                models.INSTANCE_STATUS.shutdown.value
            ]
        ).order_by('id')
        if not self.request.user.is_superuser:
            qs = qs.filter(owner=self.request.user)
        if self.q:
            qs = qs.filter(name__istartswith=self.q)
        return qs

class CloudViewSet(viewsets.ReadOnlyModelViewSet):
    queryset = models.Cloud.objects.all()
    serializer_class = serializers.CloudSerializer

class ImageViewSet(viewsets.ModelViewSet):#TODO allow specify image id when post
    filter_fields=('public',)
    http_method_names = ['get','post','delete','head','options']
    permission_classes = (IsAuthenticated,permissions.IsOwnerOrAdminOrPublicReadOnly,)
    serializer_class = serializers.ImageSerializer
    def get_queryset(self):
        queryset = models.Image.objects.filter(Q(public=True) | Q(owner=self.request.user)).order_by('-pk')
        return queryset
    def perform_create(self, serializer):
        serializer.save(owner=self.request.user)

class InstanceTemplateViewSet(viewsets.ReadOnlyModelViewSet):
    serializer_class = serializers.InstanceTemplateSerializer
    queryset = models.InstanceTemplate.objects.all().order_by('-pk')

class InstanceBlueprintViewSet(viewsets.ModelViewSet):
    filter_fields=('public',)
    permission_classes = (IsAuthenticated,permissions.IsOwnerOrAdminOrPublicReadOnly,)
    http_method_names = ['get','post','delete','head','options']
    serializer_class = serializers.InstanceBlueprintSerializer
    def get_queryset(self):
        queryset = models.InstanceBlueprint.objects.filter(Q(public=True) | Q(owner=self.request.user)).order_by('-pk')
        return queryset
    def perform_create(self, serializer):
        serializer.save(owner=self.request.user)

class InstanceViewSet(viewsets.ModelViewSet):
    filter_fields=('cluster',)
    #TODO diallow to modify specific fields
    permission_classes = (IsAuthenticated,permissions.IsOwnerOrAdmin)
    serializer_class = serializers.InstanceSerializer
    def get_queryset(self):
        queryset = models.Instance.objects.filter(owner=self.request.user).order_by('-pk')
        return queryset
    def perform_create(self, serializer):
        serializer.save(owner=self.request.user)

class InstanceOperationViewSet(viewsets.ModelViewSet):
    serializer_class = serializers.InstanceOperationSerializer
    http_method_names = ['get','post','head','options']
    def get_queryset(self):
        queryset = models.InstanceOperation.objects.filter(target__owner=self.request.user).order_by('-pk')
        return queryset

class VolumeViewSet(viewsets.ModelViewSet):
    serializer_class = serializers.VolumeSerializer
    permission_classes = (IsAuthenticated,permissions.IsOwner,)
    def get_queryset(self):
        queryset = models.Volume.objects.filter(owner=self.request.user).order_by('-pk')
        return queryset
    def perform_create(self, serializer):
        serializer.save(owner=self.request.user)

class MountViewSet(viewsets.ModelViewSet):
    serializer_class = serializers.MountSerializer
    http_method_names = ['get','post','delete','head','options']
    def get_queryset(self):
        queryset = models.Mount.objects.filter(volume__owner=self.request.user).order_by('-pk')
        return queryset



from django.contrib import admin
from django.db.models import Q, fields
from django.utils.html import format_html
from django.utils.timezone import now
from daa.admin import AutoModelAdmin
from clouds.utils import get_formated_url, get_url
from user.models import Balance
from .models import OPERATION_STATUS
from django.db import transaction

#TODO use staticmodel of_user
def powerful_form_field_queryset_Q(db_field, request):
    if db_field.name == 'cloud':
        return Q(balance__in=request.user.balances())
    model=db_field.remote_field.model
    q=None
    if getattr(model, 'owner', False):
        q = Q(owner=request.user)
        if getattr(model, 'public', False):
            q= (q | Q(public=True))
        if getattr(model, 'enabled', False):
            q= q & Q(enabled=True)
    return q

class OwnershipModelAdmin(AutoModelAdmin):
    # def action_checkbox(self, obj):
    #     u=get_current_user()
    #     if not u.is_superuser and u!=obj.owner: return None
    #     return super().action_checkbox(obj)
    # action_checkbox.short_description=AutoModelAdmin.action_checkbox.short_description
    def has_change_permission(self, request, obj=None):
        return not obj or obj.owner==request.user
    def has_delete_permission(self, request, obj=None):
        return not obj or obj.owner==request.user or request.user.is_superuser
    def get_queryset_Q(self, request):
        return Q(owner=request.user)
    def save_model(self, request, obj, form, change):
        obj.owner = request.user
        obj.save()
    def get_form_field_queryset_Q(self, db_field, request):
        return powerful_form_field_queryset_Q(db_field, request)

class StaticModelAdmin(OwnershipModelAdmin):
    search_fields = ('name','remark')
    list_filter = ('public', 'enabled',
        ('owner', admin.RelatedOnlyFieldListFilter),
    )
    form_fields_exclude=('name',)
    def get_fields(self,request,obj):
        return ('name',) + super().get_fields(request,obj)
    def get_list_display_exclude(self, request, obj=None):
        return ('created_time','name')
    def get_list_display(self, request, obj=None):
        return ('name',)+super().get_list_display(request,obj)
    def get_queryset_Q(self, request):
        return super().get_queryset_Q(request) | (Q(public=True) & Q(enabled=True))

class OperatableAdminMixin(object):
    def action_button(self, obj, op_url):
        ops=obj.get_running_operations()
        if ops.exists(): return ops[0].operation+'ing'
        former=obj.get_former_operation()
        if former and former.status==OPERATION_STATUS.failed.value: return OPERATION_STATUS.failed.value
        if not obj.startable and not obj.stopable: return None
        if obj.stopable:
            op='poweroff'
        else:
            op='start'
        return format_html('<a href="{}" data-id={} data-op={} class="button">{}</a>'.format(op_url,obj.pk,op,op))
    def has_change_permission(self, request, obj=None):
        return False #obj.ready and super().has_change_permission(request,obj)
    def has_delete_permission(self, request, obj=None):
        return not obj or obj.owner==request.user and (obj.ready or obj.deleting) or request.user.is_superuser

class OperationAdmin(AutoModelAdmin):
    search_fields = ('script',)
    list_filter = (
        'operation',
        ('target', admin.RelatedOnlyFieldListFilter),
        'manual',
        'status',
    )
    def get_list_display(self,request,obj=None):
        return super().get_list_display(request,obj)
    def _target(self,obj):
        return get_formated_url(obj.target)
    list_display_exclude=('operation','target','tidied','created_time','batch_uuid','script','log')
    def get_queryset_Q(self, request):
        return Q(target__owner=request.user)
    def get_form_fields_exclude(self,request,obj=None):
        return ('target',) if obj else ('_target',)
    def get_list_display(self,request,obj=None):
        return ('batch','operation', '_target','short_script') + super().get_list_display(request,obj)+('short_log',)
    def get_readonly_fields(self,request,obj=None):
        return ('_target', 'batch_uuid',) + super().get_readonly_fields(request,obj)
    def get_form_field_queryset_Q(self, db_field, request):
        return powerful_form_field_queryset_Q(db_field, request)
    def has_change_permission(self, request, obj=None):
        return False
    def has_delete_permission(self, request, obj=None):
        return not obj or obj.target.owner == request.user or request.user.is_superuser
    @transaction.atomic
    def rerun(modeladmin, request, queryset):
        for op in queryset:
            op=op.__class__.objects.select_for_update().get(pk=op.pk)
            op.status=OPERATION_STATUS.running.value
            op.started_time=now()
            op.completed_time=None
            op.log=None
            op.save()
            op.execute()

    rerun.short_description = "Re-run selected operations"
    actions=[rerun]

class M2MOperationAdmin(OperationAdmin):
    def sub_operations(self,obj):
        return format_html('<br/>'.join([get_url(sub) for sub in obj.get_sub_operations()]))
    def get_readonly_fields(self, request, obj=None):
        fs=super().get_readonly_fields(request, obj)
        if obj: fs+=('sub_operations',)
        return fs


from django.contrib import admin
from . import models
from clouds.base.admin import StaticModelAdmin, OwnershipModelAdmin, OperatableAdminMixin, OperationAdmin
from django.utils.html import format_html
from django.urls import reverse
from django.shortcuts import redirect
from django import forms
from django.db.models import Q
from dal import autocomplete
from user.utils import get_current_user, get_space

@admin.register(models.DataEngine)
class DataEngineAdmin(StaticModelAdmin):
    search_fields = ('description',)+StaticModelAdmin.search_fields
    
@admin.register(models.DataSource)
class DataSourceAdmin(StaticModelAdmin):
    search_fields = ('description',)+StaticModelAdmin.search_fields
        
@admin.register(models.Dataset)
class DatasetAdmin(StaticModelAdmin):
    search_fields = ('description',)+StaticModelAdmin.search_fields
    def action(self,obj):
        return format_html('<a href="{}?dataset={}" class="button">Load</a>'.format(reverse('admin:data_datainstance_add'),obj.pk))
    extra=('action',)

@admin.register(models.DataInstance)
class DataInstanceAdmin(OwnershipModelAdmin,OperatableAdminMixin):
    class DataInstanceForm(forms.ModelForm):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.fields['cluster'].widget = forms.HiddenInput()
            self.fields['cluster'].initial = get_space().pk
        class Meta:
            model = models.DataInstance
            fields = ('__all__')
            widgets = {
                'engine': autocomplete.ModelSelect2(
                    url='datainstanceengine-autocomplete',
                    forward=['dataset','cluster']
                ),
            }
    form = DataInstanceForm
    def query(self, obj):
        if not obj.ready or not obj.cluster.ready: return None
        url=obj.cluster.portal.split('//')[-1].split(':')[0]
        return format_html('<a href="{}" type="{}" target="_blank" class="button">Query</a>'.format(url, obj.dataset.type_name))
    extra=('uri','query')#,'action'
    search_fields = ('name', 'dataset__name', 'cluster__name', 'engine__name')+OwnershipModelAdmin.search_fields
    list_filter = (
        ('dataset', admin.RelatedOnlyFieldListFilter),
        ('engine', admin.RelatedOnlyFieldListFilter),
        ('cluster', admin.RelatedOnlyFieldListFilter),
    )+OwnershipModelAdmin.list_filter
    def get_list_display_exclude(self, request, obj=None):
        if request.user.is_superuser: 
            return ()
        return ('owner','deleting')
    def get_queryset_Q(self, request):
        return (super().get_queryset_Q(request)) and Q(cluster=get_space())
    def response_add(self, request, obj, post_url_continue=None):
        url=reverse("admin:data_datainstance_changelist")
        space=get_space()
        if space: url='/space/{}'.format(space.pk)+url
        return redirect(url)

@admin.register(models.DataInstanceOperation)
class DataInstanceOperationAdmin(OperationAdmin):
    def get_list_display(self,request,obj=None):
        return super().get_list_display(request,obj)+('log',)
    def get_queryset_Q(self, request):
        return super().get_queryset_Q(request) and Q(target__cluster=get_space())


import time
import json
from uuid import uuid4
from enum import Enum
from threading import Thread
from django.db import models
from django.db.models import Q
from django.contrib.auth.models import User
from django.core.validators import MinValueValidator
from clouds.models import StaticModel, Image, Instance, Mount, INSTANCE_STATUS, INSTANCE_OPERATION, OPERATION_STATUS, InstanceBlueprint, InstanceOperation, OperatableMixin, OperationModel, Group, GroupOperation
from django.utils.functional import cached_property
from django.db import transaction
from clouds.base.models import M2MOperatableMixin, M2MOperationModel
from engines.drivers.ambari import Driver

import pkgutil
from django.conf import settings

class COMPONENT_STATUS(Enum):
    null=0 #unknown
    active=1
    block=2
    suspend=3
    stop=4
    breakdown=5
    pause=6
    instance_lost=7

class COMPONENT_TYPE(Enum):
    master="master"
    slave="slave"
    client="client"

class COMPONENT_OPERATION(Enum):
    start="start"
    stop="stop"
    # restart="restart"

class Engine(StaticModel):#TODO make Engine customizable in the ui
    uuid=models.UUIDField(auto_created=True, default=uuid4, editable=False)
    name=models.CharField(max_length=100,unique=True)
    description=models.TextField(max_length=5120,blank=True,default='')
    public=models.BooleanField(default=True, editable=False)
    def get_host(self, cluster):
        return cluster.get_instances().first()
        # hostname=self.stack_set.first().driver.get_engine_host(cluster.portal, self.name)
        # return cluster.find_instance(hostname)

import importlib
class Stack(StaticModel):
    name=models.CharField(max_length=100,unique=True,verbose_name='version')
    engines=models.ManyToManyField(Engine,blank=True)
    meta=models.TextField(max_length=5120,default="",blank=True,verbose_name='meta data in json format')
    public=models.BooleanField(default=True, editable=False)
    @cached_property
    def meta_data(self):
        return json.loads(self.meta)

class Scale(StaticModel):
    init_blueprints=models.ManyToManyField(InstanceBlueprint,related_name="initialized_scales",verbose_name='initial blueprints')
    step_blueprints=models.ManyToManyField(InstanceBlueprint,related_name="stepped_scales",blank=True,verbose_name='scale-out blueprints')
    stack=models.ForeignKey(Stack,on_delete=models.PROTECT)
    _remedy_script=models.TextField(max_length=5120,default="",blank=True,verbose_name='initial remedy script')
    _remedy_script_scale_out=models.TextField(max_length=5120,default="",blank=True,verbose_name='scale-out remedy script')
    _remedy_script_scale_in=models.TextField(max_length=5120,default="",blank=True,verbose_name='scale-in remedy script')
    auto=models.BooleanField(default=False)
    def __str__(self):
        return "{}/{}".format(
            self.name,
            'auto' if self.auto else 'manual',
        )
    @cached_property
    def remedy_script_scale_out(self):
        return "###scale-out remedy {}: {}###\n{}\n".format(
            self._meta.verbose_name,
            self.name,
            self._remedy_script_scale_out
        ) if self._remedy_script_scale_out else ""
    @cached_property
    def remedy_script_scale_in(self):
        return "###scale-in remedy {}: {}###\n{}\n".format(
            self._meta.verbose_name,
            self.name,
            self._remedy_script_scale_in
        ) if self._remedy_script_scale_in else ""
    @cached_property
    def init_size(self):
        q=0
        for ib in self.init_blueprints.all():
            q+=ib.quantity
        return q
    @cached_property
    def step_size(self):
        q=0
        for ib in self.step_blueprints.all():
            q+=ib.quantity
        return q
    @cached_property
    def available_engines(self):
        return self.stack.engines.all()
    def scale(self, owner, current_step=0, remark=None):
        step=Group(owner=owner)
        step.save()
        if not current_step:#init
            if not remark: remark='initialized from scale: {}'.format(self.name)
            step.remark=remark
            for ib in self.init_blueprints.all():
                inss=ib.launch(owner=owner, remark=remark)
                step.instances.add(*inss)
        else:
            if not remark: remark='scaled from scale: {}'.format(self.name)
            step.remark=remark
            for ib in self.step_blueprints.all():
                next_number=ib.quantity*(current_step-1)+1
                if self.init_blueprints.filter(pk=ib.pk).exists():
                    next_number+=ib.quantity
                inss=ib.launch(owner=owner, next_number=next_number, remark=remark)
                step.instances.add(*inss)
        step.remark=remark
        step.save()
        return step

def scales_of_user(self):
    excluded_blueprints=InstanceBlueprint.objects.exclude(pk__in=self.blueprints())
    return Scale.objects.exclude(
        Q(init_blueprints__in=excluded_blueprints) | Q(step_blueprints__in=excluded_blueprints)
    ).filter(enabled=True).filter(Q(public=True) | Q(owner=self)).distinct()
User.scales=scales_of_user

class Cluster(models.Model,M2MOperatableMixin):
    uuid=models.UUIDField(auto_created=True, default=uuid4, editable=False)
    name=models.CharField(max_length=50)
    scale=models.ForeignKey(Scale,on_delete=models.PROTECT)
    engines=models.ManyToManyField(Engine,blank=True)
    steps=models.ManyToManyField(Group,blank=True,editable=False)
    remedy_script_todo=models.TextField(max_length=51200,default="",blank=True)
    remark = models.CharField(blank=True,null=True,max_length=100)
    public=models.BooleanField(default=False)
    owner=models.ForeignKey(User,on_delete=models.PROTECT,editable=False)
    created_time=models.DateTimeField(auto_now_add=True)
    built_time=models.DateTimeField(blank=True, null=True, editable=False)
    status= models.PositiveIntegerField(choices=[(status.value,status.name) for status in INSTANCE_STATUS],default=INSTANCE_STATUS.building.value,editable=False)
    deleting = models.BooleanField(default=False,editable=False)
    class Meta:
        unique_together = ('name', 'owner')
    def __str__(self):
        return "{}".format(self.name)
    @staticmethod
    def get_operation_model():
        return ClusterOperation
    @property
    def operatables(self):
        return self.steps.all()
    @cached_property
    def portal(self):#TODO formalize, opt perf.
        if not self.ready: raise Exception('cluster not ready')
        mi=self.get_ready_instances().filter(hostname__contains='master1')
        if mi.exists(): return "http://"+str(mi[0].ipv4)+':8080'
        return None
    @cached_property
    def driver(self):
        return Driver(self.portal)
    def import_engine(self):
        stack=self.scale.stack
        for e in self.driver.stack_engines:
            eg, created=Engine.objects.get_or_create(
                name=e.name,
                defaults={
                    'owner': self.owner,
                    'description': e.description,
                    'remark': 'auto imported',
                    'enabled': stack.enabled
                }
            )
            stack.engines.add(eg)
    @property
    def metrics(self):
        return self.driver.metrics
    @property
    def ready(self):
        return self.built_time
    @property
    def building(self):
        return self.built_time and not self.ready
    def get_ready_steps(self):
        return self.steps.filter(built_time__isnull=False)
    def get_instances(self):
        return Instance.objects.filter(group__in=self.steps.filter()).distinct()
    def get_ready_instances(self):
        return Instance.objects.filter(group__in=self.get_ready_steps()).distinct()
    @cached_property
    def engines_unselected(self):
        return Engine.objects.all().difference(self.engines.all())
    @transaction.atomic
    def scale_one_step(self):# the only way to scale cluster
        cluster=self.__class__.objects.select_for_update().get(pk=self.pk)
        step=cluster.scale.scale(
            owner=cluster.owner,
            current_step=cluster.steps.count(),
            remark='cluster: '+cluster.name
        )
        cluster.steps.add(step)
    def delete(self, *args, **kwargs):
        if not self.ready:
            print('WARNNING: delete {} under building'.format(self._meta.verbose_name))
        operatables=self.operatables
        if operatables.exists():
            self.deleting=True
            self.save()
            for operatable in operatables:
                operatable.destroy_script_todo=self.scale.remedy_script_scale_in
                operatable.save()
                operatable.delete()
        else:
            super().delete(*args, **kwargs)

    # def start(self):
    #     return utils.ambari_service_start_all('admin','admin',self.portal)#TODO use credential args
    # def stop(self):
    #     return utils.ambari_service_stop_all('admin','admin',self.portal)
    # def add_selected_engines(self):
    #     for e in self.engines.all():
    #         #set maintaince mode instead of removing
    #         utils.ambari_service_maintenance_off('admin','admin',self.portal,e.name.upper())
    # def remove_unselected_engines(self):
    #     for e in self.engines_unselected:
    #         #set maintaince mode instead of removing
    #         utils.ambari_service_maintenance_on('admin','admin',self.portal,e.name.upper())
    # @cached_property
    # def init_size(self):
    #     size=0
    #     for ib in self.blueprint.instance_blueprints.all():
    #         size+=ib.quantity
    #     return size
    def find_instance(self,hostname):
        return self.get_instances().filter(hostname__contains=hostname).first()

def clusters_of_user(self):
    return Cluster.objects.filter(Q(public=True) | Q(owner=self))
User.clusters=clusters_of_user

def steps_of_user(self):
    return Group.objects.filter(cluster__in=self.clusters()).distinct()
User.steps=steps_of_user

class ClusterOperation(M2MOperationModel):
    target=models.ForeignKey(Cluster,on_delete=models.CASCADE)
    class Meta:
        verbose_name='cluster operation'
    @staticmethod
    def get_sub_operation_model():
        return GroupOperation

class StepOperation(GroupOperation):
    class Meta(GroupOperation.Meta):
        proxy = True
        verbose_name = 'operation'
    def __str__(self):
        return "Cluster {}/Step {}/{}/{}".format(self.cluster,self.number,self.operation,self.status)
    @cached_property
    def cluster(self):
        return self.target.cluster_set.first()
    @cached_property
    def number(self):
        steps=self.cluster.steps.all()
        l=steps.count()
        target_pk=self.target.pk
        for i in range(l):
            if steps[i].pk==target_pk:
                return i+1


import requests
import time
import json
import re
import logging, os, sys

class Driver(object):
    def __init__(self, cloud):
        self.log = logging.getLogger(os.path.basename(__file__).split('.')[0].upper())
        if not len(self.log.handlers): #prevent to add duplicated handlers
            self.log.setLevel(logging.DEBUG)
            ch = logging.StreamHandler(sys.stdout)
            ch.setLevel(logging.DEBUG)
            formatter = logging.Formatter('%(asctime)s %(name)s %(levelname)s - %(message)s')
            ch.setFormatter(formatter)
            self.log.addHandler(ch)
        self._cloud=cloud
        self._credential=cloud.platform_credential
        self._headers={'Content-Type': 'application/json'}
        self.endpoint=self._credential['endpoint']
        token=requests.post(
            url=self.endpoint+'/v3/auth/tokens',
            headers=self._headers,
            data=json.dumps({
                "auth":{
                    "identity": {
                        "methods": ["password"],
                        "password": {"user": {"domain": {"id": "default"},"name": self._credential['username'],"password": self._credential['password']}}
                    },
                    "scope": {
                        "project": {"domain": {"id": "default"},"name": self._credential['project_name']}
                    }
                }
            })
        ).headers['X-Subject-Token']
        self._token=token
        self._headers['X-Auth-Token']=token
        self.project_id=requests.get(url=self.endpoint+'/v3/projects?name='+self._credential['project_name'], headers=self._headers).json()['projects'][0]['id']
        self.tenant_endpoint='/v2/'+self.project_id
        self.instances=InstanceManager(self)
        self.volumes=VolumeManager(self)
        self.images=ImageManager(self)
        self.flavors=FlavorManager(self)
        self.keypairs=KeyManager()
    def _request(self,url,method=requests.get,data=None,headers=None,retry_when_response_unexpected_strings=None,retry_until_response_expected_strings=None):
        if not headers:
            headers=self._headers
        mustend = time.time() + 1800
        while time.time() < mustend:
            res = method(self.endpoint+url,headers=headers,data=json.dumps(data))
            self.log.info('REQUEST: token/{}, {} {}, status/{}'.format(self._token, method.__name__, url, res.status_code))
            if retry_when_response_unexpected_strings:
                ss=list(filter(lambda x: x in res.text, retry_when_response_unexpected_strings))
                if ss:
                    self.log.info('REQUEST RETRING, AS UNEXPECTED RESPONSE STRING "{}" appeared in response: "{}"!'.format(ss[0], res.text))
                    time.sleep(5)
                    continue
            if retry_until_response_expected_strings:
                ss=list(filter(lambda x: x in res.text, retry_until_response_expected_strings))
                if not ss:
                    self.log.info('REQUEST RETRING, AS EXPECTED RESPONSE STRING "{}"  didn\'t appear in response: "{}"!'.format(' or '.join(retry_until_response_expected_strings), res.text))
                    time.sleep(5)
                    continue
            break
        if not res.ok:
            raise Exception(res.status_code, res.text)
        try:
            if res.text: return res.json()
        except Exception as e:
            print(type(e), e)
    def _get(self, url, retry_when_response_unexpected_strings=None, retry_until_response_expected_strings=None):
        return self._request(url,retry_when_response_unexpected_strings=retry_when_response_unexpected_strings, retry_until_response_expected_strings=retry_until_response_expected_strings)
    def _tenant_get(self, url, retry_when_response_unexpected_strings=None, retry_until_response_expected_strings=None):
        return self._request(self.tenant_endpoint+url, retry_when_response_unexpected_strings=retry_when_response_unexpected_strings, retry_until_response_expected_strings=retry_until_response_expected_strings)
    def _create(self, url, data, retry_when_response_unexpected_strings=None, retry_until_response_expected_strings=None):
        return self._request(url,method=requests.post,data=data,retry_when_response_unexpected_strings=retry_when_response_unexpected_strings, retry_until_response_expected_strings=retry_until_response_expected_strings)
    def _tenant_create(self, url, data, retry_when_response_unexpected_strings=None, retry_until_response_expected_strings=None):
        return self._create(self.tenant_endpoint+url,data=data,retry_when_response_unexpected_strings=retry_when_response_unexpected_strings, retry_until_response_expected_strings=retry_until_response_expected_strings)
    def _delete(self, url, retry_when_response_unexpected_strings=None, retry_until_response_expected_strings=None):
        try:
            return self._request(url,method=requests.delete,retry_when_response_unexpected_strings=retry_when_response_unexpected_strings, retry_until_response_expected_strings=retry_until_response_expected_strings)
        except Exception as e:
            if e.args[0]!=404:
                raise e
    def _tenant_delete(self, url, retry_when_response_unexpected_strings=None, retry_until_response_expected_strings=None):
        return self._delete(self.tenant_endpoint+url,retry_when_response_unexpected_strings=retry_when_response_unexpected_strings, retry_until_response_expected_strings=retry_until_response_expected_strings)

class KeyManager(object):
    def create(self, name, public_key):
        return None
    def delete(self, name):
        return None

class FlavorManager(object):
    def __init__(self, driver):
        self.driver=driver
    def get(self, id):
        for tpl in self.list():
            if tpl.id==id:
                return tpl
        raise Exception('unfounded flavor (id: {})!'.format(id))
    def list(self):
        flavors=[]
        for item in self.driver._tenant_get('/flavors/detail')['flavors']:
            flavors.append(Flavor(item))
        return flavors
 
class Flavor(object):
    def __init__(self, info):
        self.info=info
        self.id=info['id']
        self.name=info['name']
        self.ram=info['ram']
        self.vcpus=info['vcpus']
        self.disk=info['disk']
    def __repr__(self):
        return '{}: {}'.format(type(self).__name__, self.name)

class ImageManager(object):
    def __init__(self, driver):
        self.driver=driver
    def get(self, id):
        info=self.driver._get('/v2/images/'+id)['image']
        return Image(info)
    def _json2Images(self, json):
        images=[]
        for item in json:
            images.append(Image(item))
        return images
    def find(self, name):
        return self._json2Images(self.driver._get('/v2/images?name='+name)['images'])
    def list(self):
        return self._json2Images(self.driver._get('/v2/images')['images'])
    def delete(self, id):
        return self.driver._delete('/v2/images/'+id)

class Image(object):
    def __init__(self, info):
        self.info=info
        self.id=info['id']
        self.name=info['name']
        self.min_ram=info['min_ram']*1024
        self.min_disk=info['min_disk']
        self.created_at=info['created_at']
    def __repr__(self):
        return '{}: {}'.format(type(self).__name__, self.name)

class InstanceManager(object):
    mountable_status=['ACTIVE','SHUTDOWN']
    def __init__(self, driver):
        self.driver=driver
    def get(self, id):
        info=self.driver._tenant_get('/servers/'+id)['server']
        return Instance(self, info)
    def list(self):
        inss=[]
        for item in self.driver._tenant_get('/servers/detail')['servers']:
            inss.append(Instance(self, item))
        return inss
    def create(self, image_id, template_id, remark='packone', **kwargs):
        data={
        	"server": {
        		"name": re.sub(r'[\W,_]', '-', remark).strip('-'),
        		"imageRef": image_id,
        		"flavorRef": template_id,
                "availability_zone": self.driver._credential["nova-availability_zone"],
        		"networks": [
        			{
        				"uuid": self.driver._credential['net-id']
        			}
        		]
        	}
        }
        info=self.driver._tenant_create('/servers',data=data, retry_when_response_unexpected_strings=['is not active'])['server']
        info=self.driver._tenant_get('/servers/'+info['id'],retry_until_response_expected_strings=['ACTIVE'])['server']
        ins=Instance(self, info)
        from ..utils import SSH
        ssh=SSH(ins.addresses['provider'][0]['addr'],'root',password=info['adminPass'])
        pub=self.driver._cloud._public_key
        ssh.exec_batch('HM=~{}; cd $HM; mkdir -p .ssh; cd .ssh; echo "{}">>authorized_keys; chmod 400 authorized_keys'.format(self.driver._cloud.instance_credential_username, pub))
        ssh.close()
        return ins
    def delete(self, id):
        try:
            return self.driver._tenant_delete('/servers/'+id)
        except Exception as e:
            print(e)
    def force_delete(self, id):
        return self.delete(id)
    def get_status(self, id):
        return self.get(id).status

class Instance(object):
    def __init__(self, manager, info):
        self.info=info
        self.manager=manager
        self.id=info['id']
        self.addresses={}
        for addr in info['addresses']:
            self.addresses['provider']=info['addresses'][addr] #TODO remove 'provider' key
            break
        self.name=info['name']
        self.status=info['status']
        self.created=info['created']
    def __repr__(self):
        return '{}: {}-{}'.format(type(self).__name__, self.name, self.addresses['provider'][0]['addr'] if 'provider' in self.addresses else self.addresses)
    def get_console_url(self, type):
        return self.manager.driver._tenant_create('/servers/{}/action'.format(self.id), data={"os-getVNCConsole": { "type": type}})
    def create_image(self, image_name):
        return self.manager.driver._tenant_create('/servers/{}/action'.format(self.id), data={"createImage":{"name":image_name,"metadata":{"operator_id":self.manager.driver.project_id}}},retry_when_response_unexpected_strings=['conflictingRequest'])
    def start(self):
        return self.manager.driver._tenant_create('/servers/{}/action'.format(self.id), data={"os-start": None})
    def reboot(self):
        return self.manager.driver._tenant_create('/servers/{}/action'.format(self.id), data={"reboot": {"type": "SOFT"}})
    def stop(self):
        return self.manager.driver._tenant_create('/servers/{}/action'.format(self.id), data={"os-stop": None})

class VolumeManager(object):
    def __init__(self, driver):
        self.driver=driver
    def get(self, id):
        info=self.driver._tenant_get('/volumes/'+id)['volume']
        return Volume(self, info)
    def list(self):
        vols=[]
        for item in self.driver._tenant_get('/volumes/detail')['volumes']:
            vols.append(Volume(self, item))
        return vols
    def create(self, size, remark='packone'):
        data={"volume": {"name": remark,"size": size, 'availability_zone': self.driver._credential['cinder-availability_zone']}}
        info=self.driver._tenant_create('/volumes',data=data)['volume']
        info=self.driver._tenant_get('/volumes/'+info['id'], retry_until_response_expected_strings=['available'])['volume']
        return Volume(self, info)
    def delete(self, id):
        self.driver._tenant_get('/volumes/'+id, retry_until_response_expected_strings = ['available', 'error'])['volume']
        return self.driver._tenant_delete('/volumes/'+id)
    def mount(self, volume_id, instance_id):
        data={"volumeAttachment": {"volumeId": volume_id}}
        self.driver._tenant_create('/servers/'+instance_id+'/os-volume_attachments',data=data)
        info = self.driver._tenant_get('/volumes/'+volume_id,retry_until_response_expected_strings=['in-use'])['volume']
        return Volume(self, info)
    def unmount(self, volume_id, instance_id):
        self.driver._tenant_get('/servers/'+instance_id,retry_until_response_expected_strings=['SHUTOFF','404'])
        return self.driver._tenant_delete('/servers/'+instance_id+'/os-volume_attachments/'+volume_id)

class Volume(object):
    def __init__(self, manager, info):
        self.info=info
        self.manager=manager
        self.id=info['id']
        self.size=info['size']
        self.name=info['name']
        self.created_at=info['created_at']
        self.attachments=info['attachments']
        self.status=info['status']
    def __repr__(self):
        return '{}: {}-{} GB'.format(type(self).__name__, self.name, self.size)


import time
from uuid import uuid4
from enum import Enum
from threading import Thread
from django.db import models
from django.db.models import Q
from django.db import transaction
from django.dispatch import Signal
from django.contrib.auth.models import User
from django.core.validators import MinValueValidator
from django.utils.functional import cached_property
from .base.models import INSTANCE_OPERATION, INSTANCE_STATUS, OPERATION_STATUS, StaticModel, OperatableMixin, OperationModel, M2MOperatableMixin, M2MOperationModel
from django.utils.timezone import now
from django.db.utils import IntegrityError

import pkgutil
from django.conf import settings
drivers=[]
for importer, modname, ispkg in pkgutil.iter_modules((settings.BASE_DIR+'/clouds/drivers',)):
    driver='clouds.drivers.{}'.format(modname)
    drivers.append((driver,driver))

bootstraped = Signal(providing_args=["instance","name"])

import importlib, json
class Cloud(StaticModel):
    _driver=models.CharField(max_length=50,choices=drivers)
    _platform_credential=models.TextField(max_length=5120,blank=True,null=True)
    instance_credential_username=models.CharField(max_length=50, default='root')
    instance_credential_private_key=models.TextField(max_length=2048,blank=True,null=True,editable=False)
    hosts=models.TextField(max_length=5120,default='127.0.0.1 localhost localhost.localdomain localhost4 localhost4.localdomain4\n::1 localhost localhost.localdomain localhost6 localhost6.localdomain6',blank=True,null=True)
    owner=models.ForeignKey(User,on_delete=models.PROTECT,editable=False,verbose_name='admin')
    @cached_property
    def _key_name(self):
        import re
        return 'PackOne_'+re.sub('\W','-',self.name)
    @cached_property
    def _public_key(self):
        from .utils import get_pub_key
        return get_pub_key(self.instance_credential_private_key)+' '+self._key_name
    @cached_property
    def driver_module(self):
        return importlib.import_module(self._driver)
    @cached_property
    def driver(self):
        return self.driver_module.Driver(self)
    @cached_property
    def platform_credential(self):
        return json.loads(self._platform_credential)
    def import_image(self, cloud_images=None):
        if not cloud_images: cloud_images=self.driver.images.list()
        for img in cloud_images:
            name=img.name
            id=img.id
            if Image.objects.filter(cloud=self, access_id=id).exists(): continue
            if Image.objects.filter(cloud=self, name=name).exists():
                name='{}-{}'.format(name,id)
            Image(
                name=name,
                cloud=self,
                access_id = id,
                min_ram=img.min_ram,
                min_disk=img.min_disk,
                hostname='packone',
                owner=self.owner,
                remark='auto imported',
                enabled=self.enabled,
                public=self.public,
                created_time=img.created_at
            ).save()
    def import_template(self):
        for tpl in self.driver.flavors.list():
            id=tpl.id
            if InstanceTemplate.objects.filter(cloud=self, access_id=id).exists(): continue
            InstanceTemplate(
                access_id=id,
                name=tpl.name,
                ram=tpl.ram,
                vcpus=tpl.vcpus,
                disk=tpl.disk,
                cloud=self,
                owner=self.owner,
                remark='auto imported',
                enabled=self.enabled,
                public=self.public
            ).save()
    def bootstrap(self):#TODO diss rely on centos7
        img=self.image_set.filter(name__iregex=r'CentOS.*?7.*?GenericCloud').order_by('-created_time').first()
        if not img: raise Exception('image CentOS.*?7.*?GenericCloud is required!')
        flavor=self.instancetemplate_set.filter(ram__gte=max(8192,img.min_ram),vcpus__gte=2,disk__gte=max(30,img.min_disk)).order_by('ram', 'vcpus', 'disk').first()
        if not flavor: raise Exception('a flavor(ram>=8192, vcpus>=2) is required!')
        from .utils import remedy_image_ambari_agent, remedy_image_ambari_server
        image_ambari_agent, created=Image.objects.get_or_create(
            name='packone-bootstrap-ambari-agent',
            parent=img,
            access_id=img.access_id,
            min_ram=img.min_ram,
            min_disk=img.min_disk,
            cloud=self,
            owner=self.owner,
            remark='auto created',
            _remedy_script=remedy_image_ambari_agent()
        )
        image_ambari_server, created=Image.objects.get_or_create(
            name='packone-bootstrap-ambari-server',
            parent=image_ambari_agent,
            access_id=img.access_id,
            min_ram=img.min_ram,
            min_disk=img.min_disk,
            cloud=self,
            owner=self.owner,
            remark='auto created',
            _remedy_script=remedy_image_ambari_server()
        )
        image_master1, created=Image.objects.get_or_create(
            name='packone-bootstrap-master1',
            parent=image_ambari_server,
            access_id=img.access_id,
            min_ram=img.min_ram,
            min_disk=img.min_disk,
            hostname='master1.packone',
            owner=self.owner,
            remark='auto created',
            cloud=self
        )
        image_master2, created=Image.objects.get_or_create(
            name='packone-bootstrap-master2',
            parent=image_ambari_agent,
            access_id=img.access_id,
            min_ram=img.min_ram,
            min_disk=img.min_disk,
            hostname='master2.packone',
            owner=self.owner,
            remark='auto created',
            cloud=self
        )
        image_slave, created=Image.objects.get_or_create(
            name='packone-bootstrap-slave',
            parent=image_ambari_agent,
            access_id=img.access_id,
            min_ram=img.min_ram,
            min_disk=img.min_disk,
            hostname='slave.packone',
            owner=self.owner,
            remark='auto created',
            cloud=self
        )
        blueprint_master1, created=InstanceBlueprint.objects.get_or_create(
            name='packone-bootstap-master1',
            cloud=self,
            template=flavor,
            image=image_master1,
            remark='auto created',
            owner=self.owner
        )
        blueprint_master2, created=InstanceBlueprint.objects.get_or_create(
            name='packone-bootstap-master2',
            cloud=self,
            template=flavor,
            image=image_master2,
            remark='auto created',
            owner=self.owner
        )
        blueprint_slave, created=InstanceBlueprint.objects.get_or_create(
            name='packone-bootstap-slave',
            cloud=self,
            template=flavor,
            image=image_slave,
            remark='auto created',
            owner=self.owner
        )
        bootstraped.send(sender=self.__class__, instance=self, name='bootstraped')
            
def clouds_of_user(self):
    return Cloud.objects.filter(
        balance__in=self.balances(),
        enabled=True,
    ).filter(Q(public=True) | Q(owner=self)).distinct()
User.clouds=clouds_of_user

class Gate(StaticModel):
    name=models.CharField(max_length=500)
    cloud=models.ForeignKey(Cloud,on_delete=models.CASCADE)
    endpoint=models.GenericIPAddressField(protocol='IPv4',verbose_name='wireguard endpoint',help_text="public IP")
    port=models.PositiveIntegerField(default=51820,validators=[MinValueValidator(2)],verbose_name='wireguard port',help_text="public port")
    credential=models.CharField(max_length=5000, blank=True, null=True,verbose_name='wireguard credential',help_text="public key")
    networks=models.CharField(max_length=500,verbose_name='routed networks',help_text="seperated with ;")
    class Meta:
        unique_together = ('cloud', 'name')

class Image(StaticModel):
    name=models.CharField(max_length=500)
    cloud=models.ForeignKey(Cloud,on_delete=models.CASCADE)
    access_id = models.CharField(max_length=50,verbose_name="actual id on the cloud")
    hostname=models.CharField(max_length=50,default='packone')
    parent=models.ForeignKey("self",on_delete=models.CASCADE,blank=True,null=True)
    min_ram=models.PositiveIntegerField(default=1024,validators=[MinValueValidator(256)])
    min_disk=models.PositiveIntegerField(default=30,validators=[MinValueValidator(1)])
    protected=models.BooleanField(default=True,editable=False)
    class Meta:
        unique_together = ('cloud', 'name')
    @cached_property
    def remedy_script(self):
        s="###remedy image: {}###\n{}\n".format(self.name,self._remedy_script) if self._remedy_script else ""
        if self.parent:
            s=self.parent.remedy_script+s
        return s
    def launch(self, template, owner, remedy_script='', number=1, remark=None):
        hostname=self.hostname
        if number>1:
            parts=hostname.split('.')
            hostname='.'.join((parts[0]+str(number),'.'.join(parts[1:])))
        ins=Instance(
            cloud=self.cloud,
            template=template,
            image=self,
            hostname=hostname,
            remedy_script_todo=remedy_script,
            owner=owner,
            remark=remark
        )
        ins.save()
        return ins

class InstanceTemplate(StaticModel):#TODO support root volume resize
    name=models.CharField(max_length=500)
    cloud=models.ForeignKey(Cloud,on_delete=models.CASCADE)
    access_id = models.CharField(max_length=50,verbose_name="actual id on the cloud")
    vcpus=models.PositiveIntegerField(default=1, validators=[MinValueValidator(1)])
    ram=models.PositiveIntegerField(default=512,validators=[MinValueValidator(256)])
    disk=models.PositiveIntegerField(default=30,validators=[MinValueValidator(1)])
    class Meta:
        verbose_name = "flavor"
        unique_together = ('cloud', 'name')
    def __str__(self):
        return "{}/vcpus:{},ram:{},disk:{}".format(self.name,self.vcpus,self.ram,self.disk)

class InstanceBlueprint(StaticModel):
    name=models.CharField(max_length=500)
    cloud=models.ForeignKey(Cloud,on_delete=models.CASCADE)
    template=models.ForeignKey(InstanceTemplate,on_delete=models.CASCADE)
    image=models.ForeignKey(Image,on_delete=models.CASCADE,related_name="instance_blueprints")
    volume_capacity=models.IntegerField(validators=[MinValueValidator(0)],default=0)
    volume_mount_point=models.CharField(default="/data",max_length=100)
    quantity=models.PositiveIntegerField(default=1, validators=[MinValueValidator(1)])
    class Meta:
        verbose_name = "blueprint"
        unique_together = ('cloud', 'name')
    def __str__(self):
        return "{}/{}/{}".format(self.image,self.template,self.quantity)
    @transaction.atomic
    def launch(self, owner, next_number=1, remark=None):
        if not remark: remark='launched from instance_blueprint: {}'.format(self)
        inss=[]
        for number in range(next_number,next_number+self.quantity):
            ins=self.image.launch(
                template=self.template,
                owner=owner,
                remedy_script=self.remedy_script,
                number=number,
                remark=remark
            )
            inss.append(ins)
            if self.volume_capacity:
                volume=Volume(
                    cloud=self.cloud,
                    capacity=self.volume_capacity,
                    owner=owner,
                    remark=remark
                )
                volume.save()
                Mount(
                    point=self.volume_mount_point,
                    instance=ins,
                    volume=volume
                ).save()
        return inss

def blueprints_of_user(self):
    return InstanceBlueprint.objects.filter(
        cloud__in=self.clouds(),
        enabled=True
    ).filter(Q(public=True) | Q(owner=self))
User.blueprints=blueprints_of_user

#TODO class VLan

monitored = Signal(providing_args=["instance","name"])

#TODO change model status field to char
def to_status_value(status):
    if status=='ERROR':
        status = INSTANCE_STATUS.failure.value
    elif status=='SHUTOFF':
        status = INSTANCE_STATUS.shutdown.value
    else:
        try:
            status = INSTANCE_STATUS[status.lower()].value
        except Exception as e:
            print(e)
            status = INSTANCE_STATUS.null.value
    return status

class Instance(models.Model,OperatableMixin):
    ipv4=models.GenericIPAddressField(protocol='IPv4',blank=True,null=True,editable=False)
    ipv6=models.GenericIPAddressField(protocol='IPv6',blank=True,null=True,editable=False)
    cloud=models.ForeignKey(Cloud,on_delete=models.PROTECT)
    uuid=models.UUIDField(unique=True,auto_created=False,null=True,editable=False)
    image=models.ForeignKey(Image,on_delete=models.PROTECT)
    template=models.ForeignKey(InstanceTemplate,on_delete=models.PROTECT)
    hostname=models.CharField(max_length=50,blank=True,null=True)#TODO not null, unique and templated
    remedy_script_todo=models.TextField(max_length=51200,default="",blank=True)
    created_time=models.DateTimeField(auto_now_add=True)
    built_time=models.DateTimeField(blank=True,null=True,editable=False)
    remark = models.CharField(blank=True,null=True,max_length=1000)
    owner=models.ForeignKey(User,on_delete=models.PROTECT,editable=False)
    status= models.PositiveIntegerField(choices=[(status.value,status.name) for status in INSTANCE_STATUS],default=INSTANCE_STATUS.building.value,editable=False)
    deleting = models.BooleanField(default=False,editable=False)
    def __str__(self):
        return "{}/{}/{}".format(self.image.name,self.template.name,self.ipv4)
    @staticmethod
    def get_operation_model():
        return InstanceOperation
    @property
    def ready(self):
        return self.uuid
    @property
    def building(self):
        return self.built_time and not self.ready
    @cached_property
    def vnc_url(self):
        return self.cloud.driver.instances.get(str(self.uuid)).get_console_url('novnc')['console']['url']
    @cached_property
    def hosts_record(self):
        if not self.ready: raise Exception('instance not ready')
        hostname_parts=self.hostname.split('.')
        hostnames=' '.join(('.'.join(hostname_parts[0:pi+1]) for pi in range(0,len(hostname_parts))))		
        return self.ipv4+' '+hostnames
    @property
    def mountable(self):
        if INSTANCE_STATUS(self.status).name.upper() in self.cloud.driver_module.InstanceManager.mountable_status: return True
        if self.status==INSTANCE_STATUS.building.value and self.ready: return True
        return False
    @property
    def umountable(self):
        return self.mountable
    def monitor(self,notify=True):
        if not self.ready: raise Exception('instance not ready')
        status = self.cloud.driver.instances.get_status(str(self.uuid))
        if self.__class__.objects.filter(pk=self.pk).update(status=to_status_value(status)):
            self.refresh_from_db()
            if notify: monitored.send(sender=self.__class__, instance=self, name='monitored')
    def get_or_create_image(self, image_name=None):
        if not image_name: image_name=self.hostname
        images=list(filter(lambda x: x.name==image_name, self.cloud.driver.images.list()))
        if images:
            created=False
        else:
            self.cloud.driver.instances.get(str(self.uuid)).create_image(image_name)
            created=True
        images=list(filter(lambda x: x.name==image_name, self.cloud.driver.images.list()))
        self.cloud.import_image(images)
        image = Image.objects.get(cloud=self.cloud, name=image_name)
        image.hostname=self.hostname
        image.protected=False
        image.save()
        return image, created
    @property
    def credential(self):
        return self.owner.profile_set.filter(enabled=True).first().credential
    @cached_property
    def password(self):
        return self.credential.ssh_passwd
    @cached_property
    def public_key(self):
        return self.credential.ssh_public_key
    def set_public_key(self):
        if self.public_key:
            self.remedy("echo '{}'>>/root/.ssh/authorized_keys".format(self.public_key),manual=False)
    def set_password(self):
        if self.password:
            self.remedy("echo 'root:{}' | chpasswd".format(self.password),manual=False)
    
class VOLUME_STATUS(Enum):#greater value means worse status
    null=0 #unknown
    available=1
    mounted=2
    building=3 #executing remedy_script

class Volume(models.Model):
    cloud=models.ForeignKey(Cloud,on_delete=models.PROTECT)
    uuid=models.UUIDField(unique=True,auto_created=False,null=True,editable=False)
    capacity=models.IntegerField()
    created_time=models.DateTimeField(auto_now_add=True)
    built_time=models.DateTimeField(blank=True,null=True,editable=False)
    remark = models.CharField(blank=True,null=True,max_length=100)
    owner=models.ForeignKey(User,on_delete=models.PROTECT,editable=False)
    status= models.PositiveIntegerField(choices=[(status.value,status.name) for status in VOLUME_STATUS],default=VOLUME_STATUS.building.value,editable=False)
    deleting = models.BooleanField(default=False,editable=False)
    def __str__(self):
        return "{}/{}/{}".format(self.cloud.name,self.capacity,str(self.uuid).split('-')[0])
    @property
    def ready(self):
        return self.uuid
    @property
    def building(self):
        return self.built_time and not self.ready
    @property
    def mountable(self):
        #TODO get mount condition from driver
        return self.status == VOLUME_STATUS.available.value
    @property
    def umountable(self):
        return self.status == VOLUME_STATUS.mounted.value
    def get_running_operations(self):
        return Mount.objects.filter(
            volume=self,
            dev__isnull=True,
            completed_time__isnull=False
        ).order_by('id')

#TODO add same-cloud check
class Mount(models.Model):
    volume=models.OneToOneField(Volume,on_delete=models.PROTECT)
    instance=models.ForeignKey(Instance,on_delete=models.PROTECT)
    dev=models.CharField(max_length=100,null=True,editable=False)
    point=models.CharField(default="/data",max_length=200)
    created_time=models.DateTimeField(auto_now_add=True)
    completed_time=models.DateTimeField(blank=True,null=True,editable=False)
    class Meta:
        unique_together = ('instance', 'point')
        default_permissions=('add', 'delete', 'view')
    def __str__(self):
        return "{}/{}/{}".format(self.instance,self.dev,self.point)
    @property
    def ready(self):
        return self.dev
    @property
    def executing(self):
        return self.completed_time and not self.ready

executed = Signal(providing_args=["instance","name"])

class InstanceOperation(OperationModel):
    target=models.ForeignKey(Instance,on_delete=models.CASCADE)
    log=models.TextField(max_length=51200,null=True,editable=False)
    def execute(self):
        def perform(self=self):
            from . import utils
            import traceback
            if self.operation!=INSTANCE_OPERATION.remedy.value:
                try:
                    ins=self.target.cloud.driver.instances.get(str(self.target.uuid))
                    if self.operation==INSTANCE_OPERATION.start.value:
                        if self.target.cloud.driver.instances.get_status(str(self.target.uuid))==INSTANCE_STATUS.active.value:
                            output='skip start op: already running'
                        else:
                            output=ins.start()
                    elif self.operation==INSTANCE_OPERATION.reboot.value:
                        output=ins.reboot()
                    elif self.operation in [INSTANCE_OPERATION.poweroff.value, INSTANCE_OPERATION.shutdown.value]:
                        if self.target.cloud.driver.instances.get_status(str(self.target.uuid)) in [INSTANCE_STATUS.poweroff.value, INSTANCE_STATUS.shutdown.value]:
                            output='skip stop op: already stoped'
                        else:
                            output=ins.stop()
                    if self.is_boot:
                        cloud=self.target.cloud
                        utils.SSH(self.target.ipv4,cloud.instance_credential_username,private_key=cloud.instance_credential_private_key).close() #to wait booting finished
                except Exception as e:
                    print(e)
                    # traceback.print_tb(e.__traceback__)
                    self.status=OPERATION_STATUS.failed.value
                    self.log='EXCEPTION MESSAGE:\n'+str(e)
                else:
                    self.status=OPERATION_STATUS.success.value
                    self.log=output
            else:
                if not self.script:
                    self.log='EXCEPTION MESSAGE:\nblank remedy script'
                    self.status=OPERATION_STATUS.failed.value
                else:
                    try:
                        cloud=self.target.cloud
                        ssh=utils.SSH(self.target.ipv4,cloud.instance_credential_username,private_key=cloud.instance_credential_private_key)
                        out, err = ssh.exec_batch(self.script)
                        self.log=out+err
                        self.status=OPERATION_STATUS.failed.value if err else OPERATION_STATUS.success.value
                    except Exception as e:
                        traceback.print_tb(e.__traceback__)
                        self.log='EXCEPTION MESSAGE:\n'+str(e)
                        self.status=OPERATION_STATUS.failed.value
            self.completed_time=now()
            if self.ignore_error and self.status==OPERATION_STATUS.failed.value:
                self.status=OPERATION_STATUS.success.value
            try:
                self.save()
            except IntegrityError as e:
                print(e)
            executed.send(sender=InstanceOperation, instance=self, name='executed')
        Thread(target=perform).start()

destroyed = Signal(providing_args=["instance","name"])

class Group(models.Model,M2MOperatableMixin):
    uuid=models.UUIDField(auto_created=True, default=uuid4, editable=False)
    instances=models.ManyToManyField(Instance,blank=True,editable=False)
    remedy_script_todo=models.TextField(max_length=51200,default="",blank=True)
    destroy_script_todo=models.TextField(max_length=51200,default="",blank=True)
    hosts=models.TextField(max_length=5120,default="",blank=True,editable=False)
    remark = models.CharField(blank=True,null=True,max_length=100)
    owner=models.ForeignKey(User,on_delete=models.PROTECT,editable=False)
    created_time=models.DateTimeField(auto_now_add=True)
    built_time=models.DateTimeField(blank=True, null=True, editable=False)
    status= models.PositiveIntegerField(choices=[(status.value,status.name) for status in INSTANCE_STATUS],default=INSTANCE_STATUS.building.value,editable=False)
    deleting = models.BooleanField(default=False,editable=False)
    class Meta:
        ordering = ['pk']
    def __str__(self):
        return self.long_id
    @property
    def long_id(self):
        return str(self.uuid).split('-')[0]
    @staticmethod
    def get_operation_model():
        return GroupOperation
    @property
    def operatables(self):
        return self.instances.all()
    @property
    def ready(self):
        return self.built_time
    @property
    def building(self):
        return self.built_time and not self.ready
    @property
    def mounts(self):
        return Mount.objects.filter(instance__in=self.instances.all())
    @transaction.atomic
    def delete(self, *args, **kwargs):
        if not self.ready:
            print('WARNNING: delete {} under building'.format(self._meta.verbose_name))
        elif self.destroy_script_todo:
            self.remedy(self.destroy_script_todo)
        operatables=self.operatables
        if operatables.exists():
            self.deleting=True
            self.save()
            for operatable in operatables:
                operatable.deleting=True
                operatable.save()
                for m in operatable.mount_set.select_related('volume').select_for_update():
                    if m.ready:
                        m.volume.deleting=True
                        m.volume.save()
                    else:
                        m.delete()
                InstanceOperation(
                    target=operatable,
                    operation=INSTANCE_OPERATION.poweroff.value,
                    ignore_error=True
                ).save()
                time.sleep(0.1)
        else:
            destroyed.send(sender=Group, instance=self, name='destroyed')
            super().delete(*args, **kwargs)

class GroupOperation(M2MOperationModel):
    target=models.ForeignKey(Group,on_delete=models.CASCADE)
    def __str__(self):
        return "{}({}/{})".format(self.batch,self.operation,self.status)
    class Meta(M2MOperationModel.Meta):
        verbose_name = "group operation"
    @staticmethod
    def get_sub_operation_model():
        return InstanceOperation

class InstanceMetric(models.Model):
    instance=models.ForeignKey(Instance,on_delete=models.CASCADE)
    uuid=models.UUIDField(unique=True,auto_created=False,null=True,editable=False)
    metric=models.TextField(max_length=51200,default="",blank=True)# in json format
    created_time=models.DateTimeField(auto_now_add=True)
    remark = models.CharField(blank=True,null=True,max_length=100)


from django.contrib import admin
from django.db.models import Q
from dal import autocomplete
from django import forms
from django.utils.html import format_html
from django.urls import reverse
from django.shortcuts import redirect
from django.contrib import messages
from user.models import Balance
from user.utils import get_current_user, get_space
from .utils import get_url, get_formated_url
from .base.admin import AutoModelAdmin, StaticModelAdmin, OwnershipModelAdmin, OperatableAdminMixin, OperationAdmin, M2MOperationAdmin, powerful_form_field_queryset_Q
from . import models
    
@admin.register(models.Cloud)
class CloudAdmin(StaticModelAdmin):
    list_filter = ('_driver',)+StaticModelAdmin.list_filter
    def get_exclude(self, request, obj=None):
        if obj and obj.owner!=request.user:
            return ('_platform_credential', 'instance_credential_private_key')
        return ()
    def import_image(modeladmin, request, queryset):
        for cloud in queryset:
            cloud.import_image()
    import_image.short_description = "refresh images from selected clouds"
    def import_template(modeladmin, request, queryset):
        for cloud in queryset:
            cloud.import_template()
    import_template.short_description = "refresh templates from selected clouds"
    def bootstrap(modeladmin, request, queryset):
        for cloud in queryset:
            cloud.bootstrap()
    bootstrap.short_description = "bootstrap selected clouds"
    actions = [import_image,import_template, bootstrap]

class CloudStaticModelAdmin(StaticModelAdmin):
    search_fields = ('cloud__name',)+StaticModelAdmin.search_fields
    list_filter = (
        ('cloud', admin.RelatedOnlyFieldListFilter),
    )+StaticModelAdmin.list_filter
    def get_queryset_Q(self, request):
        return (super().get_queryset_Q(request)) & (Q(cloud__in=request.user.clouds()) | Q(cloud__owner=request.user))
    def has_delete_permission(self, request, obj=None):
        return not obj or obj.owner==request.user or obj.cloud.owner == request.user or request.user.is_superuser

@admin.register(models.Gate)
class GateAdmin(CloudStaticModelAdmin):
    pass
            
@admin.register(models.Image)
class ImageAdmin(CloudStaticModelAdmin):
    class ImageForm(forms.ModelForm):
        class Meta:
            model = models.Image
            fields = ('__all__')
            widgets = {
                'parent': autocomplete.ModelSelect2(
                    url='image-autocomplete',
                    forward=['cloud']
                )
            }
    form = ImageForm
    def clone(self,obj):
        return format_html('<a href="{}?access_id={image.access_id}&cloud={image.cloud.pk}&parent={image.pk}&min_ram={image.min_ram}&min_disk={image.min_disk}" class="button">Clone</a>'.format(reverse('admin:clouds_image_add'),image=obj))
    def launch(self,obj):
        return format_html('<a href="{}?cloud={image.cloud.pk}&image={image.pk}" class="button">Launch</a>'.format(reverse('admin:clouds_instance_add'),image=obj)) 
    extra=('clone','launch')
    def get_list_display_exclude(self, request, obj=None):
        return ('access_id',)+super().get_list_display_exclude(request,obj)

@admin.register(models.InstanceTemplate)
class InstanceTemplateAdmin(CloudStaticModelAdmin):
    def get_list_display_exclude(self, request, obj=None):
        return ('access_id',)+super().get_list_display_exclude(request,obj)
    def has_add_permission(self, request, obj=None):
        if models.Cloud.objects.filter(owner=request.user).exists(): return True
        return False
    def get_form_field_queryset_Q(self, db_field, request):
        return Q(owner=request.user)

#TODO use css and js to move drop down list actions to list and disable operate non-public items
@admin.register(models.InstanceBlueprint)
class InstanceBlueprintAdmin(CloudStaticModelAdmin):
    search_fields = ('template__name', 'image__name')+CloudStaticModelAdmin.search_fields
    list_filter = (
        ('template', admin.RelatedOnlyFieldListFilter),
        ('image', admin.RelatedOnlyFieldListFilter),
    )+CloudStaticModelAdmin.list_filter
    #TODO filter by cloud
    def launch(modeladmin, request, queryset):
        for ib in queryset:
            ib.launch(request.user)
    launch.short_description = "Launch resources from selected blueprints"
    actions = [launch]

@admin.register(models.Instance)
class InstanceAdmin(OwnershipModelAdmin,OperatableAdminMixin):
    class InstanceForm(forms.ModelForm):
        class Meta:
            model = models.Instance
            fields = ('__all__')
            widgets = {
                'image': autocomplete.ModelSelect2(
                    url='image-autocomplete',
                    forward=['cloud']
                ),
                'template': autocomplete.ModelSelect2(
                    url='instancetemplate-autocomplete',
                    forward=['cloud','image']
                )
            }
    form = InstanceForm
    search_fields = ('cloud__name','cloud__name','template__name', 'image__name', 'hostname', 'ipv4', 'ipv6', 'remark')
    list_filter = (
        ('cloud', admin.RelatedOnlyFieldListFilter),
        ('template', admin.RelatedOnlyFieldListFilter),
        ('image', admin.RelatedOnlyFieldListFilter),
        'status',
    )
    def toggle_power(modeladmin, request, queryset):
        for ins in queryset:
            models.InstanceOperation(
                target=ins,
                ignore_error=True,
                operation=models.INSTANCE_OPERATION.poweroff.value if ins.status==models.INSTANCE_STATUS.active.value else models.INSTANCE_OPERATION.start.value
            ).save()
            
    toggle_power.short_description = "toggle power"
    def VNC(modeladmin, request, queryset):
        for ins in queryset:
            return redirect(ins.vnc_url)
    VNC.short_description = "VNC"
    def get_or_create_image(modeladmin, request, queryset):
        for ins in queryset:
            image, created = ins.get_or_create_image()
        return redirect(reverse('admin:clouds_image_change',args=(image.pk,)))
    get_or_create_image.short_description = "get or create image with the its hostname"
    actions = [VNC, toggle_power, get_or_create_image]
    def get_readonly_fields(self,request,obj=None):
        fs=super().get_readonly_fields(request,obj)
        if obj: return ('image', 'template',) + fs
        return fs
    def get_queryset_Q(self, request):
        return (super().get_queryset_Q(request) | Q(cloud__in=models.Cloud.objects.filter(owner=request.user))) & Q(group__cluster=get_space())
    # def has_delete_permission(self, request, obj=None):
    #     return not obj or obj.owner==request.user and (obj.ready or obj.deleting) or obj.cloud.owner == request.user
    
@admin.register(models.Volume)
class VolumeAdmin(OwnershipModelAdmin):
    def mounted_to(self,obj):
        return obj.mount.instance.ipv4, obj.mount.point
    def action(self,obj):
        if obj.deleting:
            if not get_current_user().is_superuser and obj.cloud.owner!=get_current_user(): 
                return 'deleting'
        ops=obj.get_running_operations()
        if ops.exists(): return 'mounting'
        if obj.umountable and obj.mount.instance.umountable:
            return format_html('<a href="{}" umount class="button">Umount</a>'.format(reverse('mount-detail',kwargs={'pk':obj.mount.pk})))
        if obj.mountable:
            return format_html('<a href="{}?volume={}" class="button">Mount</a>'.format(reverse('admin:clouds_mount_add'),obj.pk))
    search_fields = ('cloud__name','cloud__name','capacity','remark')
    list_filter = (('cloud', admin.RelatedOnlyFieldListFilter),'status')
    extra=('mounted_to','action')
    def get_list_display_exclude(self, request, obj=None):
        if request.user.is_superuser or models.Cloud.objects.filter(owner=request.user).exists(): 
            return ()
        return ('owner','deleting')
    def get_queryset_Q(self, request):
        return super().get_queryset_Q(request) | Q(cloud__in=models.Cloud.objects.filter(owner=request.user))
    def has_change_permission(self, request, obj=None):
        return False #obj.ready and super().has_change_permission(request,obj)
    # def has_delete_permission(self, request, obj=None):
    #     return not obj or obj.owner==request.user and (obj.ready or obj.deleting) or obj.cloud.owner == request.user
        
@admin.register(models.Mount)
class MountAdmin(AutoModelAdmin):
    class MountForm(forms.ModelForm):
        class Meta:
            model = models.Mount
            fields = ('__all__')
            widgets = {
                'instance': autocomplete.ModelSelect2(
                    url='mountinstance-autocomplete',
                    forward=['volume']
                ),
            }
    form = MountForm
    search_fields = ['instance__ipv4']
    def get_queryset_Q(self, request):
        return Q(volume__owner=request.user) | Q(volume__cloud__in=models.Cloud.objects.filter(owner=request.user))
    def has_change_permission(self, request, obj=None):
        return False
    def has_module_permission(self, request):
        if not request.user.is_authenticated: return False
        if request.user.is_superuser: return True
        if models.Cloud.objects.filter(owner=request.user).exists(): return True
        return False
    def response_add(self, request, obj, post_url_continue=None):
        return redirect(reverse("admin:clouds_volume_changelist"))
    def delete_view(self, request, object_id, extra_context=None):
        m=models.Mount.objects.get(pk=object_id)
        if not m.instance.umountable and request.user!=m.instance.cloud.owner:
            messages.error(request, 'not umountable instance.')
            return redirect('../..')
        return super().delete_view(request, object_id, extra_context)
    def changelist_view(self, request, extra_context=None):
        response = super().changelist_view(request, extra_context)
        if request.method == "POST" and getattr(response,'context_data',False):
            if "deletable_objects" in response.context_data:
                not_umountables=[]
                for m in response.context_data["queryset"]:
                    if not m.instance.umountable and request.user!=m.instance.cloud.owner:
                        not_umountables.append(m.instance.ipv4)
                if not_umountables:
                    messages.error(request, str(not_umountables)+' are not in umountable status')
                    return redirect('.')
        return response
    def get_form_field_queryset_Q(self, db_field, request):
        return powerful_form_field_queryset_Q(db_field, request)
    def has_change_permission(self, request, obj=None):
        return False #not obj or obj.ready and obj.volume.owner == request.user
    # def has_delete_permission(self, request, obj=None):
    #     return not obj or obj.volume.cloud.owner == request.user or (obj.instance.deleting or obj.ready) and obj.volume.owner == request.user

@admin.register(models.InstanceOperation)
class InstanceOperationAdmin(OperationAdmin):
    def get_queryset_Q(self, request):
        from user.utils import get_space
        return (super().get_queryset_Q(request)|Q(target__cloud__in=models.Cloud.objects.filter(owner=request.user))) & Q(target__group__cluster=get_space()) & ~Q(status=models.OPERATION_STATUS.success.value)
    def has_delete_permission(self, request, obj=None):
        return super().has_delete_permission(request, obj) or obj.target.cloud.owner == request.user

@admin.register(models.Group)
class GroupAdmin(OwnershipModelAdmin,OperatableAdminMixin):
    def get_list_display_exclude(self, request, obj=None):
        if request.user.is_superuser: return ()
        return ('owner','deleting',)
    def destroy(modeladmin, request, queryset):
        for group in queryset:
            group.delete()
    destroy.short_description = "Destroy selected groups"
    actions=[destroy]
    def has_add_permission(self, request, obj=None):
        return False
    def has_change_permission(self, request, obj=None):
        return False
    def has_delete_permission(self, request, obj=None):
        return False
    def has_module_permission(self, request):
        return False


from threading import Thread
import time 
from django.core.management.base import BaseCommand, CommandError
from engines.models import Cluster, COMPONENT_STATUS
class Command(BaseCommand):
    help = 'monitor engines periodly'
    def handle(self, *args, **options):
        while True:
            for c in Cluster.objects.all():
                c.status=COMPONENT_STATUS.running.value
                for e in c.engines.all():
                    if e.status(c)!=COMPONENT_STATUS.running.value:
                        c.status=COMPONENT_STATUS.stop.value
                        break
                c.save()
            time.sleep(300)


from django.contrib.auth.models import User
from rest_framework import viewsets
from rest_framework.decorators import api_view
from rest_framework.response import Response
from . import models
from . import serializers

def _get_enabled_profile(request):
    return models.Profile.objects.filter(account=request.user,enabled=True)[0]
    
@api_view(['GET'])
def user_info(request):
    profile = _get_enabled_profile(request)
    roles=[]
    if profile.account.is_active:
        roles.append('user')
    if profile.account.is_superuser:
        roles.append('admin')
    return Response({
        "organization": profile.organization,
        "name": profile.account.first_name if profile.account.first_name else profile.account.username,
        "roles": roles,
        "avatar": profile.avatar
        })

class AccountViewSet(viewsets.ModelViewSet):
    """
    API endpoint that allows users to be viewed or edited.
    """
    serializer_class = serializers.AccountSerializer
    http_method_names = ['get','post','delete','head','options']
    def get_queryset(self):
        queryset = User.objects.filter(pk=self.request.user.pk).order_by('-date_joined')
        return queryset

class ProfileViewSet(viewsets.ModelViewSet):
    serializer_class = serializers.ProfileSerializer
    def get_queryset(self):
        queryset = models.Profile.objects.filter(account=self.request.user).order_by('-pk')
        return queryset
    def perform_create(self, serializer):
        serializer.save(account=self.request.user)

class BalanceViewSet(viewsets.ReadOnlyModelViewSet):#TODO do we need add permission_classes = (IsAuthenticated, permissions.IsOwner,)
    serializer_class = serializers.BalanceSerializer
    def get_queryset(self):
        queryset = models.Balance.objects.filter(profile__account=self.request.user)
        return queryset

class CredentialViewSet(viewsets.ModelViewSet):
    serializer_class = serializers.CredentialSerializer
    def get_queryset(self):
        queryset = models.Credential.objects.filter(profile=_get_enabled_profile(self.request))
        return queryset
    def perform_create(self, serializer):
        serializer.save(profile=_get_enabled_profile(self.request))


from rest_framework import permissions

class IsOwnerOrAdminOrPublicReadOnly(permissions.BasePermission):
    """
    Custom permission to only allow owners of an object to edit it.
    """

    def has_object_permission(self, request, view, obj):
        # Read permissions are allowed to any request,
        # so we'll always allow GET, HEAD or OPTIONS requests.
        if request.method in permissions.SAFE_METHODS and obj.public:
            return True

        # Write permissions are only allowed to the owner of the snippet.
        return obj.owner == request.user or request.user.is_superuser

class IsOwner(permissions.BasePermission):
    def has_object_permission(self, request, view, obj):
        return obj.owner == request.user

class IsOwnerOrAdmin(permissions.BasePermission):
    def has_object_permission(self, request, view, obj):
        return obj.owner == request.user or request.user.is_superuser


from django.urls import re_path, include
from rest_framework import routers
from . import views

ROUTER = routers.DefaultRouter()
ROUTER.register(r'engines', views.EngineViewSet, basename='engine')
ROUTER.register(r'scales', views.ScaleViewSet, basename='scale')
ROUTER.register(r'clusters', views.ClusterViewSet, basename='cluster')
ROUTER.register(r'clusteroperations', views.ClusterOperationViewSet, basename='clusteroperation')

urlpatterns = [
    re_path(r'^', include(ROUTER.urls)),
]



import json
import urllib
import httplib2
from collections import defaultdict

from django.contrib.auth.models import User
from django.contrib.auth import login as auth_login
from django.conf import settings
from django.http import HttpResponse
from django.http import HttpResponseRedirect
from django.urls import reverse
from django.db import transaction


def _do_post(host, url, args={}):
    '''
    发送POST请求
    '''
    args = urllib.parse.urlencode(args)
    headers = {"Content-type": "application/x-www-form-urlencoded", "Accept": "text/plain"}
    conn = httplib2.HTTPSConnectionWithTimeout(host)
    conn.request('POST', url, args, headers)

    res = conn.getresponse()
    if res.status == 200:
        data = json.loads(res.read().decode())
        return data
    return False

def escience_callback(request):
    """
    平台科技网回调接口
    登录回调操作：（带code参数）
        1、根据code获取第三方账户用户信息，根据用户信息查找与之绑定的平台用户或者创建对应的用户
        2、设置用户登录
        3、跳转到首页
    """
    if settings.DEBUG: print('平台科技网回调接口')
    code = request.GET.get('code', None)
    if code:
        args = {
            "client_id": settings.ESCIENCE_APP_KEY,
            "client_secret": settings.ESCIENCE_APP_SECRET,
            "grant_type": "authorization_code",
            "redirect_uri": settings.ESCIENCE_CALLBACK,
            "code":code
        }
        
        data = _do_post(settings.ESCIENCE_LOGIN_HOST, settings.ESCIENCE_TOKEN_URL, args)

        escience_login_success = False
        user_data = defaultdict(lambda:None)
        if data and 'userInfo' in data:
            try:
                for k, v in json.loads(data.get('userInfo')).items():
                    user_data[k] = v
            except Exception as e:
                import traceback
                traceback.print_exc()
                print('[CSNET ERROR]', data.get('userInfo'), e)
            if user_data['cstnetIdStatus'] == 'active':
                escience_login_success = True

        if escience_login_success:
            user = None

            # 根据科技网邮箱查找用户
            user_objs = User.objects.filter(email=user_data["cstnetId"])            
            if user_objs.exists():
                user = user_objs.first()

            # 科技网账户和科技网邮箱都没有绑定，则注册新用户
            if user is None:
                user, created = User.objects.get_or_create(username=user_data['cstnetId'],
                        email=user_data['cstnetId'], last_name=user_data['truename'], is_staff=True)

                # 注册失败，报错
                if not created:
                    raise Exception("login fail!")

            # 登录
            if user:
                auth_login(request, user)
                return HttpResponseRedirect(reverse('scispace_index'))
    return HttpResponseRedirect(reverse('scispace_login'))



from uuid import uuid4
from enum import Enum
from django.db import models
from django.contrib.auth.models import User
from django.core.validators import MinValueValidator
from django.utils.functional import cached_property
from engines.models import Engine, COMPONENT_OPERATION
from clouds.base.models import StaticModel, OperationModel, OperatableMixin
from clouds.models import OPERATION_STATUS, INSTANCE_STATUS
from engines.models import Cluster#TODO use worldwide namespace when upload to pip
from django.core.validators import RegexValidator

class DataSource(StaticModel):
    uuid=models.UUIDField(auto_created=True, default=uuid4, editable=False)
    uri=models.CharField(max_length=200,help_text='specify the protocol and address used to sync this remote data source.')
    sync_interval=models.PositiveIntegerField(default=3600, validators=[MinValueValidator(1)],verbose_name="sync interval in seconds")
    description=models.TextField(max_length=5120)
    
class DATASET_TYPE(Enum):
    relational=0
    raw=1
    freetext=2
    semistructed=3
    event=4
    graph=5
    star=6

#TODO security exposure risk: data load histoy cmd in the relying cluster
class Dataset(StaticModel):
    uuid=models.UUIDField(auto_created=True, default=uuid4, editable=False)
    source=models.ForeignKey(DataSource,on_delete=models.PROTECT,blank=True,null=True)
    uri=models.CharField(max_length=2000)
    type=models.PositiveIntegerField(blank=True,null=True,choices=[(type.value,type.name) for type in DATASET_TYPE])
    size=models.PositiveIntegerField(default=1, validators=[MinValueValidator(1)])
    description=models.TextField(max_length=5120)
    @property
    def type_name(self):
        return DATASET_TYPE(self.type).name

class DataEngine(StaticModel):
    uuid=models.UUIDField(auto_created=True, default=uuid4, editable=False)
    type=models.PositiveIntegerField(choices=[(type.value,type.name) for type in DATASET_TYPE])
    engine=models.ForeignKey(Engine,on_delete=models.PROTECT)
    uri_prefix=models.CharField(max_length=2000, default='', blank=True)
    description=models.TextField(max_length=5120)

class INSTANCE_STATUS(Enum):
    loading=0 #unknown
    running=1
    stop=2

class DataInstance(models.Model,OperatableMixin):
    uuid=models.UUIDField(auto_created=True, default=uuid4, editable=False)
    name=models.CharField(max_length=50, validators=[RegexValidator(r'[a-z0-9]*',message='muste be lower-case alphanumeric',)])
    dataset=models.ForeignKey(Dataset,on_delete=models.PROTECT)
    cluster=models.ForeignKey(Cluster,on_delete=models.PROTECT)
    engine=models.ForeignKey(DataEngine,on_delete=models.PROTECT)#TODO:its uri should be the prefix of the final uri
    remedy_script_todo=models.TextField(max_length=51200,default="",blank=True)
    created_time=models.DateTimeField(auto_now_add=True)
    built_time=models.DateTimeField(blank=True,null=True,editable=False)
    owner=models.ForeignKey(User,on_delete=models.PROTECT,editable=False)
    status= models.PositiveIntegerField(choices=[(status.value,status.name) for status in INSTANCE_STATUS],default=INSTANCE_STATUS.loading.value,editable=False)
    remark = models.CharField(blank=True,null=True,max_length=100)
    deleting = models.BooleanField(default=False,editable=False)
    class Meta:
        unique_together = (('name','cluster','owner'),('engine','name',))
    def __str__(self):
        return "{}".format(self.name)
    @cached_property
    def entry_host(self):#the host to execute load dataset operations
        if self.engine.name=='EventDB': return self.cluster.find_instance('master1')
        if self.engine.name=='gStore': return self.cluster.find_instance('master1')
        if self.engine.name=='AstroServ': return self.cluster.find_instance('master1')
        return self.engine.engine.get_host(self.cluster)
    @cached_property
    def uri_suffix(self):#suffix of the final uri, the only approach to access this data instance.
        return self.name.replace(' ','_')
    @cached_property
    def uri(self):#suffix of the final uri, the only approach to access this data instance.
        return self.engine.uri_prefix.format(instance=self.entry_host)+self.uri_suffix
    @property
    def startable(self):
        if self.status == INSTANCE_STATUS.loading.value: return True#TODO stop
        return False
    @staticmethod
    def get_operation_model():
        return DataInstanceOperation
    @property
    def ready(self):
        return self.built_time
    @property
    def building(self):
        return self.built_time and not self.ready
    @property
    def host(self):
        return self.built_time and not self.ready
    # @property
    # def uri_total(self):#TODO opt. perf.
    #     endpoints=[]
    #     for ins in self.cluster.instances.filter(image__in=self.engine.component.images.all()):
    #         endpoints.append(ins.hostname+"://"+self.engine.endpoint)
    #     return [endpoint+"://"+self.uri_suffix for endpoint in endpoints]
    # @property
    # def uri_alive(self):
    #     endpoints=[]
    #     for ins in self.space.pilot.cluster.instance_set.filter(image__in=self.engine.component.images.all(),status=clouds.models.INSTANCE_STATUS.running.value):
    #         endpoints.append(ins.hostname+"://"+self.engine.endpoint)
    #     return [endpoint+"://"+self.uri_suffix for endpoint in endpoints]
    # @property
    # def uri_elected(self):
    #     if not self.uri_alive:
    #         return None
    #     return self.uri_alive[0]#TODO make load banlance
    # @property
    # def status_name(self):
    #     return COMPONENT_STATUS(self.status).name
    # def update_status(self):
    #     pes=self.engine.component.engine_set.filter(id__in=self.space.pilot.engines.all())
    #     self.status=pes[0].status(self.space.pilot)
    #     self.save()

class DataInstanceOperation(OperationModel):
    #TODO add custom validator for operation 'start' on 'running' instance. model.clean()?
    #TODO add custom validator for allowed operation.
    operation=models.CharField(max_length=50,choices=[(op.value,op.name) for op in COMPONENT_OPERATION],default=COMPONENT_OPERATION.start.value)
    target=models.ForeignKey(DataInstance,on_delete=models.CASCADE)
    log=models.TextField(max_length=51200,null=True,editable=False)
    def execute(self):
        self.completed_time=datetime.datetime.now()
        self.save()
        #TODO add data instance monitoring
        # print('{}:{}'.format(instance.target.ipv4,result))


# coding=utf-8

import json

from django.urls import reverse
from django.http import Http404, HttpResponseRedirect
from django.template.response import TemplateResponse
from django.http.response import JsonResponse 
from django.contrib.auth.decorators import login_required
from django.views.decorators.csrf import csrf_exempt

from scispace.utils import get_cluster_list, get_cluster_info, add_cluster, operate_cluster
from scispace.utils import get_cluster_instances, operate_cluster_instance, get_cluster_instance_info
from scispace.utils import get_dataset_list, get_dataset_info
from scispace.utils import get_data_instance_list, get_data_instance_info, add_data_instance, delete_data_instance
from scispace.utils import get_step_operation_list, get_step_operation_info
from scispace.utils import get_data_instance_operation_list, get_data_instance_operation_info
from scispace.utils import get_cluster_data_engine_list
from scispace.utils import get_scale_list, get_available_engines
from scispace.utils import get_data_metrics, get_hosts_metrics

LOGIN_URL = "/login/"

@login_required(login_url=LOGIN_URL)
def index(request):
	return HttpResponseRedirect(reverse("scispace_cluster_list"))


@login_required(login_url=LOGIN_URL)
def cluster_list(request):
	dic = {}
	dic["clusters"] = get_cluster_list(request.user)
	return TemplateResponse(request, "cluster_list.html", dic)


@login_required(login_url=LOGIN_URL)
def cluster_add(request):
	user_id = request.user.id
	dic = {}
	dic["scales"] = get_scale_list()
	# dic["required_engines"] = ["HDFS","HIVE","MAPREDUCE2","YARN","ZEPPELIN","HUE"]

	if request.method == "POST":
		name = request.POST.get("name")
		scale = int(request.POST.get("scale"))
		engines = request.POST.getlist("engines")
		engines = [ int(e) for e in engines]
		remedy_script_todo = request.POST.get("remedy_script_todo")
		public = request.POST.get("public")
		if public == "0":
			public = False
		else:
			public = True
		remark = request.POST.get("remark")
		add_cluster(user_id, name, scale, engines, public, remedy_script_todo, remark)	
		return HttpResponseRedirect(reverse("scispace_cluster_list"))

	scale_engines = {}  # not use ajax
	for sc in dic["scales"]:
		engines = get_available_engines(sc["id"])
		scale_engines[sc["id"]] = engines
	dic["scale_engines"] = json.dumps(scale_engines)
	return TemplateResponse(request, "cluster_add.html", dic)


@csrf_exempt
@login_required(login_url=LOGIN_URL)
def cluster_operate_ajax(request):
	"""
	cluster operations:
	("start","reboot","shutdown","poweroff","remedy")
    ("scale_in", "scale_out")
    ("destroy", )
	"""
	dic = {"res": True, "err":None}
	cluster_id = request.POST.get("cluster_id")
	op = request.POST.get("operation")
	if cluster_id.isdecimal():
		cluster_id = int(cluster_id)
		dic["res"] = operate_cluster(request.user, cluster_id, op)
	else:
		dic["res"] = False
		dic["err"] = "Invalid ID"
	return JsonResponse(dic)

@csrf_exempt
@login_required(login_url=LOGIN_URL)
def cluster_get_info_ajax(request):
	"""
	get cluster status
	"""
	dic = {"res": True, "info":None, "err":None}
	cluster_id = request.GET.get("cluster_id")	
	if cluster_id.isdecimal():
		cluster_id = int(cluster_id)
		cluster_info = get_cluster_info(request.user, cluster_id)
		if not cluster_info:
			raise Http404
		dic["info"] = {"status":cluster_info["status"], "status_name":cluster_info["status_name"]}
	else:
		dic["res"] = False
		dic["err"] = "Invalid ID"
	return JsonResponse(dic)


@login_required(login_url=LOGIN_URL)
def scale_engines_ajax(request):
	scale_id = int(request.GET.get("scale"))
	dic = {"res": True, "err":None, "list":[]}
	dic["list"] = get_available_engines(scale_id)
	return JsonResponse(dic)


@login_required(login_url=LOGIN_URL)
def cluster_info(request, c_id):
	dic = {}
	dic["cluster"] = get_cluster_info(request.user, c_id)	
	if not dic["cluster"]:
		raise Http404 
		
	dic["hosts"] = get_cluster_instances(request.user, c_id)
	if dic["cluster"]["portal"]:
		dic["pipeline_api_url"] = "%s/piflow-web/api/flowList" %(dic["cluster"]["portal"].replace("8080","6001").rstrip("/"))
		dic["notebook_api_url"] = "%s/api/notebook" %(dic["cluster"]["portal"].replace("8080","9995").rstrip("/"))
	else:
		dic["pipeline_api_url"] = "#"
		dic["notebook_api_url"] = "#"
	return TemplateResponse(request, "cluster_info.html", dic)

@login_required(login_url=LOGIN_URL)
def cluster_instance_list(request, c_id):
	dic = {}
	dic["cluster"] = get_cluster_info(request.user, c_id)	
	if not dic["cluster"]:
		raise Http404 
	dic["instances"] = get_cluster_instances(request.user, c_id)
	return TemplateResponse(request, "cluster_instance_list.html", dic)


@csrf_exempt
@login_required(login_url=LOGIN_URL)
def cluster_instance_operate_ajax(request, c_id):
	"""
	 operations:
	("toggle","delete",)
	"""
	dic = {"res": True, "err":None}
	instance_id = request.POST.get("instance_id")
	op = request.POST.get("operation")
	if instance_id.isdecimal():
		instance_id = int(instance_id)
		dic["res"] = operate_cluster_instance(request.user, c_id, instance_id, op)
	else:
		dic["res"] = False
		dic["err"] = "Invalid ID"
	return JsonResponse(dic)


@csrf_exempt
@login_required(login_url=LOGIN_URL)
def cluster_instance_get_info_ajax(request, c_id):
	"""
	get cluster instance status
	"""
	dic = {"res": True, "info":None, "err":None}
	instance_id = request.GET.get("instance_id")
	require_vnc = request.GET.get("require_vnc")
	if require_vnc == "true":
		require_vnc = True
	else:
		require_vnc = False
	if instance_id.isdecimal():
		instance_id = int(instance_id)
		instance_info = get_cluster_instance_info(request.user, instance_id,require_vnc=require_vnc)	
		if not instance_info:
			raise Http404 
		dic["info"] = {"status":instance_info["status"], "status_name":instance_info["status_name"], "vnc_url":instance_info["vnc_url"]}
	else:
		dic["res"] = False
		dic["err"] = "Invalid ID"
	return JsonResponse(dic)


@login_required(login_url=LOGIN_URL)
def dataset_list(request, c_id):
	dic = {}
	dic["cluster"] = get_cluster_info(request.user, c_id)
	dic["datasets"] = get_dataset_list(request.user)
	return TemplateResponse(request, "dataset_list.html", dic)

@login_required(login_url=LOGIN_URL)
def dataset_info(request, c_id, dt_id):
	dic = {}
	dic["cluster"] = get_cluster_info(request.user, c_id)
	dic["dataset"] = get_dataset_info(dt_id)
	return TemplateResponse(request, "dataset_info.html", dic)

@login_required(login_url=LOGIN_URL)
def dataset_load(request, c_id, dt_id):
	pass
# 	dic = {}
# 	dic["cluster"] = get_cluster_info(request.user, c_id)
# 	dic["dataset"] = get_dataset_info(dt_id)
# 	return TemplateResponse(request, "dataset_info.html", dic)

@login_required(login_url=LOGIN_URL)
def data_instance_list(request, c_id):
	dic = {}
	dic["cluster"] = get_cluster_info(request.user, c_id)
	dic["data_instances"] = get_data_instance_list(request.user, c_id)
	return TemplateResponse(request, "data_instance_list.html", dic)

@csrf_exempt
@login_required(login_url=LOGIN_URL)
def data_instance_get_info_ajax(request, c_id):
	"""
	get cluster instance status
	"""
	dic = {"res": True, "info":None, "err":None}
	instance_id = request.GET.get("instance_id")	
	if instance_id.isdecimal():
		instance_id = int(instance_id)
		instance_info = get_data_instance_info(instance_id)
		dic["info"] = {"status":instance_info["status"], "status_name":instance_info["status_name"]}
	else:
		dic["res"] = False
		dic["err"] = "Invalid ID"
	return JsonResponse(dic)

@login_required(login_url=LOGIN_URL)
def data_pipeline(request, c_id):
	dic = {}
	dic["cluster"] = get_cluster_info(request.user, c_id)
	if dic["cluster"]["portal"]:
		dic["pipeline_url"] = "%s/piflow-web/web/flowList?first" %(dic["cluster"]["portal"].replace("8080","6001").rstrip("/"))
	else:
		dic["pipeline_url"] = "#"
	return TemplateResponse(request, "data_pipeline.html", dic)

@login_required(login_url=LOGIN_URL)
def notebook(request, c_id):
	dic = {}
	dic["cluster"] = get_cluster_info(request.user, c_id)
	if dic["cluster"]["portal"]:
		dic["notebook_url"] = dic["cluster"]["portal"].replace("8080","9995")
	else:
		dic["notebook_url"] = "#"
	return TemplateResponse(request, "notebook.html", dic)

@login_required(login_url=LOGIN_URL)
def scispace_operation_list(request, c_id):
	dic = {}
	dic["cluster"] = get_cluster_info(request.user, c_id)
	dic["operations"] = get_step_operation_list(c_id)
	return TemplateResponse(request, "scispace_operation_list.html", dic)

@login_required(login_url=LOGIN_URL)
def scispace_operation_info(request, c_id, op_id):
	dic = {}
	dic["cluster"] = get_cluster_info(request.user, c_id)
	dic["operation"] = get_step_operation_info(c_id, op_id)
	return TemplateResponse(request, "scispace_operation_info.html", dic)


@login_required(login_url=LOGIN_URL)
def data_instance_operation_list(request, c_id):
	dic = {}
	dic["cluster"] = get_cluster_info(request.user, c_id)
	dic["operations"] = get_data_instance_operation_list(c_id)
	return TemplateResponse(request, "data_instance_operation_list.html", dic)



@login_required(login_url=LOGIN_URL)
def data_instance_operation_info(request, c_id, op_id):
	dic = {}
	dic["cluster"] = get_cluster_info(request.user, c_id)
	dic["operation"] = get_data_instance_operation_info(c_id,op_id)
	return TemplateResponse(request, "data_instance_operation_info.html", dic)


@login_required(login_url=LOGIN_URL)
def data_instance_add(request, c_id):
	user_id = request.user.id
	dic = {}
	dic["cluster"] = get_cluster_info(request.user, c_id)

	if request.method == "POST":
		name = request.POST.get("name")
		dataset_id = request.POST.get("dataset_id")
		data_engine_id = request.POST.get("data_engine_id")
		remedy_script_todo = request.POST.get("remedy_script_todo")
		remark = request.POST.get("remark")
		add_data_instance(user_id, c_id, name, dataset_id, data_engine_id, remedy_script_todo, remark)

		return HttpResponseRedirect(reverse("scispace_data_instance_list", args=(c_id,)))
	
	dic["datasets"] = get_dataset_list(request.user)
	dic["dataset_id"] = request.GET.get("dataset")
	if dic["dataset_id"]:
		dic["dataset_id"] = int(dic["dataset_id"])

	data_engines = {}  # not use ajax
	for dt in dic["datasets"]:
		engines = get_cluster_data_engine_list(c_id, dt["id"])
		data_engines[dt["id"]] = [{"id":e["id"],"name":e["engine_name"]} for e in engines]
	dic["data_engines"] = json.dumps(data_engines)

	return TemplateResponse(request, "data_instance_add.html", dic)


@login_required(login_url=LOGIN_URL)
def data_instance_delete(request, c_id):
	user_id = request.user.id
	dic = {}
	dic["cluster"] = get_cluster_info(request.user, c_id)

	if request.method == "POST":
		data_instance_id = request.POST.get("data_instance_id")
		data_instance = get_data_instance_info(data_instance_id)
		if not data_instance or data_instance["owner"].id != request.user.id:
			raise Http404
		delete_data_instance(request.user, data_instance_id)
		return HttpResponseRedirect(reverse("scispace_data_instance_list", args=(c_id,)))
	
	data_instance_id = request.GET.get("dt_id")
	if data_instance_id:
		data_instance_id = int(data_instance_id)

	data_instance = get_data_instance_info(data_instance_id)
	if not data_instance or data_instance["owner"].id != request.user.id:
		raise Http404

	dic["data_instance"] = data_instance
	return TemplateResponse(request, "data_instance_delete.html", dic)



@login_required(login_url=LOGIN_URL)
def data_instance_query(request, c_id, di_id):
	dic = {}
	dic["cluster"] = get_cluster_info(request.user, c_id)
	data_instance = get_data_instance_info(di_id)
	dic["query_url"] = data_instance["query_url"]
	dic["data_instance"] = data_instance
	return TemplateResponse(request, "data_instance_query.html", dic)



@login_required(login_url=LOGIN_URL)
def get_data_engines_ajax(request, c_id):
	dic = {"res": True, "err":None, "list":[]}
	dataset_id = request.GET.get("dataset_id")
	dataset_id = int(dataset_id)
	engines = get_cluster_data_engine_list(c_id, dataset_id)
	dic["list"] = [{"id":e["id"],"name":e["engine_name"]} for e in engines]
	return JsonResponse(dic)


@login_required(login_url=LOGIN_URL)
def get_data_metrics_ajax(request, c_id):
	dic = {"res": True, "err":None, "metrics":[]}
	metrics = get_data_metrics(c_id)
	dic["metrics"] = metrics
	return JsonResponse(dic)


@login_required(login_url=LOGIN_URL)
def get_hosts_metrics_ajax(request, c_id):
	dic = {"res": True, "err":None, "metrics":[]}
	metrics = {}
	dic["metrics"] = get_hosts_metrics(request.user, c_id)
	return JsonResponse(dic)




# coding=utf-8

from django.conf import settings
from django.urls import include, path, re_path
from django.contrib.auth import views as auth_views

from .views.main_views import *
from .views.callback import escience_callback

escience_dic = {"escience_login_url": settings.ESCIENCE_LOGIN_URL}

urlpatterns = [
    re_path(r'^$', index, name="scispace_index"),

    re_path(r'^login/$',  auth_views.LoginView.as_view(template_name="accounts_login.html", 
                                            extra_context=escience_dic), name="scispace_login"), 
    re_path(r'^logout/$', auth_views.LogoutView.as_view(), name="scispace_logout"),
    re_path(r'^callback/escience/$', escience_callback, name="callback_escience"),


    re_path(r'clusters/$', cluster_list, name="scispace_cluster_list"),
    re_path(r'clusters/add/$', cluster_add, name="scispace_cluster_add"),
    re_path(r'clusters/operate_ajax/$', cluster_operate_ajax, name="scispace_cluster_operate_ajax"),
    re_path(r'clusters/info_ajax/$', cluster_get_info_ajax, name="scispace_get_cluster_info_ajax"),
    re_path(r'clusters/scales/engines_ajax/$', scale_engines_ajax, name="scispace_scale_engines_ajax"),
    path('clusters/<int:c_id>/', cluster_info, name="scispace_cluster_info"),
    path('clusters/<int:c_id>/instances/', cluster_instance_list, name="scispace_cluster_instance_list"),
    path('clusters/<int:c_id>/instances/operate/', cluster_instance_operate_ajax, name="scispace_cluster_instance_operate_ajax"),
    path('clusters/<int:c_id>/instances/info_ajax/', cluster_instance_get_info_ajax, name="scispace_get_cluster_instance_info_ajax"),
    path('clusters/<int:c_id>/datasets/', dataset_list, name="scispace_dataset_list"),
    # path('clusters/<int:c_id>/datasets/<int:dt_id>/', dataset_info, name="scispace_dataset_info"),
    # path('clusters/<int:c_id>/datasets/<int:dt_id>/load/', dataset_info, name="scispace_dataset_load"),
    path('clusters/<int:c_id>/datainstances/', data_instance_list, name="scispace_data_instance_list"),
    path('clusters/<int:c_id>/datainstances/add/', data_instance_add, name="scispace_data_instance_add"),
    path('clusters/<int:c_id>/datainstances/delete/', data_instance_delete, name="scispace_data_instance_delete"),
    path('clusters/<int:c_id>/datainstances/query/<int:di_id>/', data_instance_query, name="scispace_data_instance_query"),
    path('clusters/<int:c_id>/datainstances/info_ajax/', data_instance_get_info_ajax, name="scispace_data_instance_get_info_ajax"),
    path('clusters/<int:c_id>/data_pipeline/', data_pipeline, name="scispace_data_pipeline"),
    path('clusters/<int:c_id>/notebook/', notebook, name="scispace_notebook"),
    path('clusters/<int:c_id>/scispace_operations/', scispace_operation_list, name="scispace_op_list"),
    path('clusters/<int:c_id>/scispace_operations/<op_id>/', scispace_operation_info, name="scispace_op_info"),
    path('clusters/<int:c_id>/data_instance_operations/', data_instance_operation_list, name="scispace_datainstance_op_list"),
    path('clusters/<int:c_id>/data_instance_operations/<op_id>/', data_instance_operation_info, name="scispace_datainstance_op_info"),
    path('clusters/<int:c_id>/data_engines_ajax/', get_data_engines_ajax, name="scispace_get_data_engines_ajax"),

    path('clusters/<int:c_id>/data_metrics_ajax/', get_data_metrics_ajax, name="scispace_get_data_metrics_ajax"),
    path('clusters/<int:c_id>/hosts_metrics_ajax/', get_hosts_metrics_ajax, name="scispace_get_hosts_metrics_ajax"),
]


from rest_framework import serializers
from django.db.models import Q
from . import models

class CloudSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.Cloud
        fields = ('name', '_driver', 'remark')

class ImageSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.Image
        fields = '__all__'

class ImagePKField(serializers.PrimaryKeyRelatedField):
    def get_queryset(self):
        user = self.context['request'].user
        queryset = models.Image.objects.filter(Q(public=True) | Q(owner=user))
        return queryset

class InstanceTemplateSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.InstanceTemplate
        fields = '__all__'

class InstanceBlueprintSerializer(serializers.ModelSerializer):
    owner = serializers.ReadOnlyField(source='owner.username')
    image=ImagePKField()
    class Meta:
        model = models.InstanceBlueprint
        fields = '__all__'

class InstanceBlueprintPKField(serializers.PrimaryKeyRelatedField):
    def get_queryset(self):
        user = self.context['request'].user
        queryset = models.InstanceBlueprint.objects.filter(Q(public=True) | Q(owner=user))
        return queryset
        
class InstanceSerializer(serializers.ModelSerializer):
    #TODO linked droplist betweeen cloud and image
    owner = serializers.ReadOnlyField(source='owner.username')
    image=ImagePKField()
    class Meta:
        model = models.Instance
        fields = '__all__'

class InstancePKField(serializers.PrimaryKeyRelatedField):
    def get_queryset(self):
        qs = models.Instance.objects.all().order_by('-pk')
        user = self.context['request'].user
        if not user.is_superuser:
            qs = qs.filter(owner=user)
        return qs

class InstanceOperationSerializer(serializers.ModelSerializer):
    #TODO filter out running instances when perform poweroff op
    target=InstancePKField()
    class Meta:
        model = models.InstanceOperation
        fields = '__all__'

class VolumeSerializer(serializers.ModelSerializer):
    owner = serializers.ReadOnlyField(source='owner.username')
    class Meta:
        model = models.Volume
        fields = '__all__'

class VolumeUnmountedPKField(serializers.PrimaryKeyRelatedField):
    def get_queryset(self):
        user = self.context['request'].user
        qs = models.Volume.objects.filter(mount=None)
        if not user.is_superuser:
            qs = qs.filter(owner=user)
        return qs

class InstanceMountablePKField(serializers.PrimaryKeyRelatedField):#TODO enable frontend filtering based on relatedField
    def get_queryset(self):
        user = self.context['request'].user
        qs = models.Instance.objects.filter(status__in=[
            models.INSTANCE_STATUS.poweroff.value,
            models.INSTANCE_STATUS.shutdown.value
        ])
        if not user.is_superuser:
            qs = qs.filter(owner=user)
        return qs

class MountSerializer(serializers.ModelSerializer):
    volume=VolumeUnmountedPKField()
    instance=InstanceMountablePKField()
    class Meta:
        model = models.Mount
        fields = '__all__'


import os, shutil
import argparse
from time import sleep

BASE_DIR = os.path.dirname(os.path.abspath(__file__))
os.chdir(BASE_DIR)

def setup(args):
    os.system('ln -sf  /usr/bin/python3.[6,7,8] /usr/bin/python3')
    database=args.database[0] if args.database else 'packone:packone:localhost:5432:packone'
    (db_user, db_passwd, db_host, db_port, db_name)=database.split(':')
    with open("conf/settings.py") as f:
        newText=f.read().replace(
            "'USER': 'rabbit'", "'USER': '"+db_user+"'"
            ).replace(
                "'PASSWORD': 'rabbit'", "'PASSWORD': '"+db_passwd+"'"
            ).replace(
                'rabbit-db-host', db_host
            ).replace(
                '5432', db_port
            ).replace(
                'packone_new', db_name
            ).replace(
                '# STATIC_ROOT', 'STATIC_ROOT'
            ).replace(
                'os.path.join(BASE_DIR, "static")', '# os.path.join(BASE_DIR, "static")'
            )
    with open("conf/settings.py", "w") as f:
        f.write(newText)

    os.system('python3 manage.py collectstatic --noinput')
    #TODO add sql migration
    os.system('python3 manage.py migrate')
    os.system('python3 manage.py makemigrations user clouds engines data')
    os.system('python3 manage.py migrate')

    print('config packone superuser')
    print('username: admin')
    os.system('python3 manage.py createsuperuser --username admin')
    
def start(args):
    address=args.listening[0] if args.listening else '0:11001'
    os.system('uwsgi --http {address} --chdir {BASE_DIR} --ini {BASE_DIR}/conf/uwsgi.ini'.format(address=address,BASE_DIR=BASE_DIR))
    sleep(3)
    os.system('python3 -mwebbrowser http://'+address)

def stop(args):
    os.system('uwsgi --stop /var/tmp/packone.pid')

def uninstall(args):
    os.system('uwsgi --stop /var/tmp/packone.pid')
    os.system('pip uninstall pk1')
    if os.path.isfile('db.sqlite3'):
        shutil.move('db.sqlite3','/var/tmp/packone.sqlite3.old')
    shutil.rmtree(BASE_DIR)

parser = argparse.ArgumentParser(description='packone cmd line.')
subparsers = parser.add_subparsers()

parser_setup = subparsers.add_parser('setup')
parser_setup.add_argument('--database', metavar='user:password:host:port:db_name', nargs=1,
                    help='the pk1 postgresql database configuration. default: packone:packone:localhost:5432:packone')
parser_setup.set_defaults(func=setup)

parser_start = subparsers.add_parser('start')
parser_start.add_argument('--listening', metavar='ip:port', nargs=1,
                    help='the pk1 listening address. default: 0:11001')
parser_start.set_defaults(func=start)

parser_stop = subparsers.add_parser('stop')
parser_stop.set_defaults(func=stop)

parser_uninstall = subparsers.add_parser('uninstall')
parser_uninstall.set_defaults(func=uninstall)

def main():
    args = parser.parse_args()
    try: #https://stackoverflow.com/a/54161510/4444459
        args.func(args)
    except AttributeError:
        parser.parse_args(['--help'])
    


from secrets import token_urlsafe
from django.dispatch import receiver
from django.db.models.signals import pre_save, post_save, post_delete, pre_delete
from django.contrib.auth.models import User
from .import models
from time import sleep

@receiver(post_save, sender=User)
def profile_auto_create(sender, instance, created, **kwargs):
    if created:
        models.Profile(owner=instance, enabled=True, remark='auto created').save()

@receiver(post_save, sender=models.Profile)
def credential_auto_create(sender, instance, created, **kwargs):
    if created:
        models.Credential(profile=instance,ssh_passwd = token_urlsafe(15)).save()

@receiver(pre_save, sender=models.Credential)
def credential_update(sender, instance, **kwargs):
    if instance.pk:
        old = models.Credential.objects.get(pk=instance.pk)
        if old.ssh_passwd!=instance.ssh_passwd:
            for ins in instance.profile.owner.instance_set.all():
                ins.remedy("echo 'root:{}' | chpasswd".format(instance.ssh_passwd),manual=False)
        if old.ssh_public_key!=instance.ssh_public_key:
            for ins in instance.profile.owner.instance_set.all():
                ins.remedy("echo '{}'>>/root/.ssh/authorized_keys".format(instance.ssh_public_key),manual=False)

@receiver(pre_delete, sender=models.Credential)
def credential_pre_delete(sender, instance, **kwargs):
    if instance.ssh_user=='root' and instance.profile.owner.instance_set.exists():
        raise Exception('cannot delete credentials of root')

@receiver(pre_save, sender=models.Profile)
def profile_pre_save(sender, instance, **kwargs):
    if instance.enabled:
        sender.objects.filter(enabled=True,owner=instance.owner).update(enabled=False)

@receiver(post_save, sender=models.Cloud)
def add_admin_balance(sender, instance, created, **kwargs):
    if created:  
        models.Balance(
            cloud=instance,
            profile=instance.owner.profile_set.filter(enabled=True).first(),
            balance=1000000
        ).save()



from uuid import uuid4
from enum import Enum
from threading import Thread
from django.db import models
from django.db import transaction
from django.dispatch import Signal
from django.contrib.auth.models import User
from django.core.validators import MinValueValidator
from django.utils.functional import cached_property
from django.utils.timezone import now

class StaticModel(models.Model):
    name=models.CharField(max_length=100,unique=True)
    _remedy_script=models.TextField(max_length=5120,default="",blank=True)
    public=models.BooleanField(default=False)
    enabled=models.BooleanField(default=True)
    modified_time=models.DateTimeField(auto_now=True)
    created_time=models.DateTimeField(auto_now_add=True)
    remark = models.CharField(blank=True,null=True,max_length=1000)
    owner=models.ForeignKey(User,on_delete=models.PROTECT,editable=False)
    class Meta:
        abstract = True
    def __str__(self):
        return "{}".format(self.name)
    @cached_property
    def remedy_script(self):
        return "###remedy {}: {}###\n{}\n".format(
            self._meta.verbose_name,
            self.name,
            self._remedy_script
        ) if self._remedy_script else ""

#TODO class VLan

class INSTANCE_STATUS(Enum):#greater value means worse status
    null=0 #unknown
    active=1
    block=2
    suspend=3
    shutdown=4
    poweroff=5
    breakdown=6
    pause=7
    failure=8 #libvirt reserv code
    host_lost=9
    instance_lost=10
    building=11

class OperatableMixin(object):
    @property
    def startable(self):
        if self.status in [
            INSTANCE_STATUS.poweroff.value,
            INSTANCE_STATUS.shutdown.value,
        ]: return True
        return False
    @property
    def stopable(self):
        return self.status == INSTANCE_STATUS.active.value
    @transaction.atomic
    def remedy(self, script='', manual=True):
        for self in self.__class__.objects.filter(pk=self.pk).select_for_update():
            if self.remedy_script_todo:
                script=self.remedy_script_todo+'\n\n'+script
                self.remedy_script_todo=''
                self.save()
            if script:
                op=self.get_operation_model()(
                    operation=INSTANCE_OPERATION.remedy.value,
                    script=script,
                    target=self,
                    manual=manual
                )
                op.save()
                return op
            return None
    def update_remedy_script(self,script,heading=False):
        if heading:
            v=models.Value(script+'\n')
            c=models.functions.Concat(v,'remedy_script_todo')
        else:
            v=models.Value('\n'+script)
            c=models.functions.Concat('remedy_script_todo',v)
        self.__class__.objects.filter(pk=self.pk).update(remedy_script_todo=c)
    def get_ready_operations(self):
        return self.get_operation_model().objects.filter(
            started_time__isnull=True,
            completed_time__isnull=True,
            status=OPERATION_STATUS.running.value,
            target=self
        ).exclude(target__deleting=True, manual=False).order_by("pk")
    def get_running_operations(self):
        return self.get_operation_model().objects.filter(
            started_time__isnull=False,
            completed_time__isnull=True,
            target=self
        ).order_by("started_time")
    def get_next_operations(self):
        return self.get_operation_model().objects.filter(
            target=self,
            started_time__isnull=True,
            completed_time__isnull=True,
            status__in=[OPERATION_STATUS.waiting.value,OPERATION_STATUS.running.value],
        ).exclude(target__deleting=True, manual=False).order_by("pk")
    def get_former_operation(self):
        return self.get_operation_model().objects.filter(
            target=self,
            status__in=[
                OPERATION_STATUS.success.value,
                OPERATION_STATUS.failed.value
            ]
        ).order_by("-started_time").first()
        
monitored = Signal(providing_args=["instance","name"])

class VOLUME_STATUS(Enum):#greater value means worse status
    null=0 #unknown
    available=1
    mounted=2
    building=3 #executing remedy_script

class INSTANCE_OPERATION(Enum):
    start="start"
    reboot="reboot"
    shutdown="shutdown"
    poweroff="poweroff"
    remedy="remedy"

class OPERATION_STATUS(Enum):
    success="success"
    failed="failed"
    running="running"
    waiting="waiting"

#TODO allow restart faild operation
class OperationModel(models.Model):
    operation=models.CharField(max_length=50,choices=[(op.value,op.name) for op in INSTANCE_OPERATION],default=INSTANCE_OPERATION.start.value)
    batch_uuid=models.UUIDField(auto_created=True, default=uuid4, editable=False)
    script=models.TextField(max_length=5120,default='',blank=True)
    serial=models.ForeignKey("self",on_delete=models.CASCADE,blank=True,null=True,editable=False)
    created_time=models.DateTimeField(auto_now_add=True)
    started_time=models.DateTimeField(blank=True,null=True,editable=False)
    completed_time=models.DateTimeField(blank=True,null=True,editable=False)
    status=models.CharField(max_length=50,choices=[(s.value,s.name) for s in OPERATION_STATUS],default=OPERATION_STATUS.running.value,editable=False)
    tidied = models.BooleanField(default=False,editable=False)
    manual = models.BooleanField(default=True,editable=False)
    ignore_error = models.BooleanField(default=False,editable=True)
    log=models.TextField(max_length=51200,null=True,blank=True,editable=False)
    class Meta:
        verbose_name = "operation"
        abstract = True
        ordering=['-started_time','-completed_time','-pk']
    def __str__(self):
        return "{}({}/{}/{})".format(self.batch,self.target,self.operation,self.status)
    @property
    def short_script(self):
        if not self.script: return self.script
        ret=[]
        for line in self.script.split('\n'):
            ret.append(line[:50])
        return '\n'.join(ret)
    @property
    def short_log(self):
        # return self.log
        if not self.log: return self.log
        ret=[]
        for line in self.log.split('\n')[:50]:
            ret.append(line[:100])
        return '\n'.join(ret)
    @property
    def batch(self):
        return str(self.batch_uuid).split('-')[0]
    @cached_property
    def is_boot(self):
        return self.operation in (INSTANCE_OPERATION.start.value,INSTANCE_OPERATION.reboot.value)
    @cached_property
    def serial_pre(self):
        if not self.serial: return None
        pre = self.__class__.objects.filter(serial=self.serial,pk__lt=self.pk).order_by('-pk').first()
        if not pre: pre=self.serial
        return pre
    @property
    def executing(self):
        return self.started_time and not self.completed_time
    @property
    def runnable(self):
        if self.serial_pre and self.serial_pre.status!=OPERATION_STATUS.success.value:
            return False
        if self.executing: return False
        if self.manual:
            if self.status!=OPERATION_STATUS.running.value: return False
            return True
        target=self.target
        if not target.stopable and self.operation==INSTANCE_OPERATION.remedy.value: return False
        if target.get_running_operations().exists(): return False
        former=target.get_former_operation()
        if former and former.status==OPERATION_STATUS.failed.value: return False
        return True
        
class M2MOperatableMixin(OperatableMixin):
    def monitor(self):
        for operatable in self.operatables:
            Thread(target=operatable.monitor).start()

executed = Signal(providing_args=["instance","name"])

class M2MOperationModel(OperationModel):
    class Meta(OperationModel.Meta):
        abstract = True
    def get_sub_operations(self):
        return self.get_sub_operation_model().objects.filter(batch_uuid=self.batch_uuid)
    def get_status(self):
        if self.get_sub_operation_model().objects.filter(
            batch_uuid=self.batch_uuid,
            status=OPERATION_STATUS.waiting.value
        ).exists(): return OPERATION_STATUS.waiting.value
        if self.get_sub_operation_model().objects.filter(
            batch_uuid=self.batch_uuid,
            status=OPERATION_STATUS.running.value
        ).exists(): return OPERATION_STATUS.running.value
        if self.get_sub_operation_model().objects.filter(
            batch_uuid=self.batch_uuid,
            status=OPERATION_STATUS.failed.value
        ).exists(): return OPERATION_STATUS.failed.value
        return OPERATION_STATUS.success.value
    def get_remain_oprations(self):
        return self.get_sub_operation_model().objects.filter(
            batch_uuid=self.batch_uuid,
            completed_time=None
        ).order_by('started_time')
    def execute(self):
        operatables=self.target.operatables
        if not operatables.exists():
            self.status=OPERATION_STATUS.running.value
            self.completed_time=now()
            self.save()
            from clouds.models import executed
            executed.send(sender=self.__class__, instance=self, name='executed')
            return
        subops=self.get_sub_operations()
        if subops.exists():
            for op in subops:
                op.execute()
        else:
            for operatable in operatables:
                self.get_sub_operation_model()(
                    operation=self.operation,
                    batch_uuid=self.batch_uuid,
                    target=operatable,
                    tidied=True,
                    manual=self.manual,
                    ignore_error=self.ignore_error,
                    script=self.script
                ).save()


from threading import local
from engines.models import Cluster

_user = local()

class CurrentUserMiddleware(object):
    def __init__(self, get_response):
        self.get_response = get_response
    def __call__(self, request):
        _user.value = request.user
        _user.space=None
        # if not request.user.is_anonymous:# TODO a workaround to allow drf anonymous user to access current space
        parts=request.get_full_path().split('/')
        if len(parts)>2 and parts[1]=='space':
            _user.space=Cluster.objects.get(pk=parts[2])
        request.space=_user.space
        return self.get_response(request)

def get_current_user():
    return _user.value

def get_space():
    return _user.space


import coreapi
#TODO directly use libvirt and ceph

class Driver(object):
    def __init__(self, cloud):
        self._cloud=cloud
        self._credential=cloud.platform_credential
        self._client=coreapi.Client(auth=coreapi.auth.BasicAuthentication(self._credential['user'], self._credential['passwd']))
        self._schema = self._client.get(self._credential['api_endpoint'])
        self.instances=InstanceManager(self)
        self.volumes=VolumeManager(self)
        self.images=ImageManager(self)
        self.flavors=FlavorManager()
        self.keypairs=KeyManager()
    def _do_action(self, action, params):
        return self._client.action(self._schema, action, params=params)

class KeyManager(object):
    def create(self, name, public_key):
        return None
    def delete(self, name):
        return None

class FlavorManager(object):
    def list(self):
        return [
            Flavor({'id': '0', 'name': 'm1.nano', 'ram': 64, 'vcpus': 1, 'disk': 30}),
            Flavor({'id': '1', 'name': 'm1.tiny', 'ram': 512, 'vcpus': 1, 'disk': 30}),
            Flavor({'id': '2', 'name': 'm1.small', 'ram': 2048, 'vcpus': 1, 'disk': 30}),
            Flavor({'id': '3', 'name': 'm1.medium', 'ram': 4096, 'vcpus': 2, 'disk': 30}),
            Flavor({'id': '4', 'name': 'm1.large', 'ram': 8192, 'vcpus': 2, 'disk': 30}),
            Flavor({'id': '5', 'name': 'm1.xlarge', 'ram': 16384, 'vcpus': 4, 'disk': 30}),
            Flavor({'id': '6', 'name': 'm1.xxlarge', 'ram': 32768, 'vcpus': 8, 'disk': 30}),
            Flavor({'id': '7', 'name': 'm1.xxxlarge', 'ram': 65536, 'vcpus': 16, 'disk': 30}),
            Flavor({'id': '8', 'name': 'packone.m', 'ram': 8192, 'vcpus': 2, 'disk': 30}),
            Flavor({'id': '9', 'name': 'packone.m+', 'ram': 10240, 'vcpus': 2, 'disk': 30}),
            Flavor({'id': '10', 'name': 'packone.l', 'ram': 16384, 'vcpus': 4, 'disk': 30}),
            Flavor({'id': '11', 'name': 'packone.x', 'ram': 32768, 'vcpus': 8, 'disk': 30}),
        ]
    def get(self, id):
        for tpl in self.list():
            if tpl.id==id:
                return tpl
        raise Exception('unfounded flavor (id: {})!'.format(id))
 
class Flavor(object):
    def __init__(self, info):
        self.id=info['id']
        self.name=info['name']
        self.ram=info['ram']
        self.vcpus=info['vcpus']
        self.disk=info['disk']

class ImageManager(object):
    def __init__(self, driver):
        self.driver=driver
    def list(self):
        action = ["images","list"]
        params = {
            "pool_id":self.driver._credential['image_poll_id'],
        }
        infos=self.driver._do_action(action,params)
        images=[]
        for info in infos:
            images.append(Image(info))
        return images

class Image(object):
    def __init__(self, info):
        self.id=info['id']
        self.name=info['name']
        self.min_ram=256
        self.min_disk=30
        if self.name=='centos7': self.name='CentOS-7.5.1804-x86_64-GenericCloud-1809'
        from django.utils.timezone import now
        self.created_at=now()
        if self.name.startswith('0000_packone'):#TODO auto build packone images
            self.name=self.name.replace('0000_','')

class InstanceManager(object):
    mountable_status=['POWEROFF','SHUTDOWN']
    def __init__(self, driver):
        self.driver=driver
    def get(self, instance_id):
        action = ["vms","read"]
        params = {
            "vm_id": instance_id
        }
        info=self.driver._do_action(action, params)
        return Instance(self, info)
    def list(self):
        action = ["vms", "list"]
        params = {
            "group_id":self.driver._credential['group_id'],
        }
        inss=[]
        for item in self.driver._do_action(action, params):
            inss.append(Instance(self, item))
        return inss
    def create(self, image_id, template_id, remark='', **kwargs):
        flavor=self.driver.flavors.get(template_id)
        action = ["vms","create"]
        params = {
            "image_id":int(image_id),
            "vcpu":flavor.vcpus,
            "mem":flavor.ram,
            "group_id":self.driver._credential['group_id'],
            'vlan_id':self.driver._credential['vlan_id'],
            "remarks":remark,
        }
        vm_id=self.driver._do_action(action, params)
        ins=self.get(vm_id)
        ins.start()
        from ..utils import SSH
        ssh=SSH(ins.addresses['provider'][0]['addr'],self.driver._credential['instance_username'],password=self.driver._credential['instance_password'])
        pub=self.driver._cloud._public_key
        ssh.exec_batch('HM=~{}; cd $HM; mkdir -p .ssh; cd .ssh; echo "{}">>authorized_keys; chmod 400 authorized_keys'.format(self.driver._cloud.instance_credential_username, pub))
        ssh.close()
        ins._operate('shutdown') #avoid disk cache lost
        ins.stop()
        return ins
    def delete(self, instance_id):
        action = ["vms","delete"]
        params = {
            "vm_id": instance_id,
        }
        try:
            return self.driver._do_action(action, params)
        except coreapi.exceptions.ErrorMessage as e:
            if '\u865a\u62df\u673aUUID\u9519\u8bef' in e.error._data['detail']:
                print('VM UUID Mismatch')
                return True
            else:
                raise e
    def force_delete(self, instance_id):
        try:
            self.get(instance_id).stop()
        except Exception as e:
            print(e)
        return self.delete(instance_id)
    def get_status(self, instance_id):
        action = ["vms","status","list"]
        params = {
            "vm_id":instance_id
        }
        try:
            status = int(self.driver._do_action(action,params))
            if status==0: return 'NULL'
            if status==1: return 'ACTIVE'
            if status==2: return 'BLOCK'
            if status==3: return 'SUSPEND'
            if status==4: return 'SHUTDOWN'
            if status==5: return 'POWEROFF'
            if status==6: return 'BREAKDOWN'
            if status==7: return 'PAUSE'
            if status==8: return 'FAILURE'
            if status==9: return 'HOST_LOST'
            if status==10: return 'INSTANCE_LOST'
        except coreapi.exceptions.ErrorMessage as e:
            if '\u865a\u62df\u673aUUID\u9519\u8bef' in e.error._data['detail']:
                print('VM UUID Mismatch')
                return 'POWEROFF'
            else:
                raise e

class Instance(object):
    def __init__(self, manager, info):
        self.manager=manager
        self.id=info['uuid']
        self.addresses={}
        if info['ipv4']: self.addresses['provider']=[{'addr': info['ipv4']}]
        self.name=info['remarks']
        self.created=info['create_time']
    def get_console_url(self, type):
        return {
            'console': {'url': self.manager.driver._do_action(["vms","vnc","create"], {"vm_id":self.id})}
        }
    def create_image(self, image_name):
        raise Exception('create image from instance is unsupported')
    def _operate(self, op):
        action = ["vms","operations","partial_update"]
        params = {"vm_id":self.id,"op":op}
        try:
            return self.manager.driver._do_action(action, params)
        except coreapi.exceptions.ErrorMessage as e:
            if 'not running' in e.error._data['detail'] or 'already running' in e.error._data['detail']:
                print(e.error._data['detail'])
                return True
            else:
                raise e
    def start(self):
        return self._operate('start')
    def reboot(self):
        return self._operate('reboot')
    def stop(self):
        return self._operate('poweroff')

class VolumeManager(object):
    def __init__(self, driver):
        self.driver=driver
    def _operate_volume(self, volume_id, instance_id, op):
        action = ["volumes","vm","partial_update"]
        params = {
            "volume_id":volume_id,
            "vm_id":instance_id,
            "op":op
        }
        return self.driver._do_action(action, params)
    def get(self, volume_id):
        action = ["volumes","read"]
        params = {
            "volume_id":volume_id
        }
        info=self.driver._do_action(action, params)
        info['uuid']=info['id']
        if info['dev']:
            info['dev']='/dev/'+info['dev']
        return Volume(self, info)
    def list(self):
        action = ["volumes", "list"]
        params = {
            "group_id":self.driver._credential['group_id'],
            "cephpool_id":self.driver._credential['volume_poll_id'],
        }
        vols=[]
        for item in self.driver._do_action(action, params):
            vols.append(Volume(self, item))
        return vols
    def create(self, size, remark=''):
        action = ["volumes","create"]
        params = {
            "group_id":self.driver._credential['group_id'],
            "cephpool_id":self.driver._credential['volume_poll_id'],
            'size':size,
            "remarks":remark
        }
        volume_id=self.driver._do_action(action, params)
        return self.get(volume_id)
    def delete(self, volume_id):
        action = ["volumes","delete"]
        params = {
            "volume_id":volume_id,
        }
        try:
            return self.driver._do_action(action, params)
        except coreapi.exceptions.ErrorMessage as e:
            if 'CEPH\u5757ID\u6709\u8bef' in e.error._data['detail']:
                print('CEPH UUID Mismatch')
                return True
            else:
                raise e
    def mount(self, volume_id, instance_id):
        self._operate_volume(volume_id, instance_id, 'mount')
        return self.get(volume_id)
    def unmount(self, volume_id, instance_id):
        try:
            return self._operate_volume(volume_id, instance_id, 'unmount')
        except coreapi.exceptions.ErrorMessage as e:
            if '\u865a\u62df\u673aUUID\u9519\u8bef' in e.error._data['detail']:
                print('VM UUID Mismatch')
                return True
            else:
                raise e

class Volume(object):
    def __init__(self, manager, info):
        self.manager=manager
        self.id=info['uuid']
        self.size=info['size']
        self.name=info['remarks']
        self.created_at=info['create_time']
        self.attachments=[]
        if info['attach_time']:
            self.attachments.append({'device': info['dev'], 'attached_at': info['attach_time']})


from django.urls import re_path, include
from rest_framework import routers
from . import views

ROUTER = routers.DefaultRouter()
ROUTER.register(r'account', views.AccountViewSet, basename='user')
ROUTER.register(r'profiles', views.ProfileViewSet, basename='profile')
ROUTER.register(r'balance', views.BalanceViewSet, basename='balance')
ROUTER.register(r'credential', views.CredentialViewSet, basename='credential')

urlpatterns = [
    re_path(r'^', include(ROUTER.urls)),
    re_path('info/', views.user_info),
]